#!/usr/env/python

"""
lossy_flow_accumulator.py: Component to accumulate flow and calc drainage area,
while permitting gain or loss of discharge during flow.

DEJH, late 2018
"""

import sys

from landlab.components.flow_accum import (
    FlowAccumulator,
    flow_accum_bw,
    flow_accum_to_n,
)

if sys.version_info[0] >= 3:
    from inspect import signature


class LossyFlowAccumulator(FlowAccumulator):

    """Component to calculate drainage area and accumulate flow, while
    permitting dynamic loss or gain of flow downstream.

    This component is closely related to the FlowAccumulator, in that
    this is accomplished by first finding flow directions by a user-specified
    method and then calculating the drainage area and discharge. However,
    this component additionally requires the passing of a function that
    describes how discharge is lost or gained downstream,
    f(Qw, nodeID, linkID, grid). See the Examples below to see how this works
    in practice.

    Optionally, spatially variable runoff can be set either by the model grid
    field 'water__unit_flux_in' or the input variable *runoff_rate**.

    Optionally a depression finding component can be specified and flow
    directing, depression finding, and flow routing can all be accomplished
    together. Note that the DepressionFinderAndRouter is not particularly
    intelligent when running on lossy streams, and in particular, it will
    reroute flow around pits even when they are in fact not filled due to loss.

    NOTE: The perimeter nodes  NEVER contribute to the accumulating flux, even
    if the  gradients from them point inwards to the main body of the grid.
    This is because under Landlab definitions, perimeter nodes lack cells, so
    cannot accumulate any discharge.

    LossyFlowAccumulator stores as ModelGrid fields:

        -  Node array of drainage areas: *'drainage_area'*
        -  Node array of discharges: *'surface_water__discharge'*
        -  Node array of discharge loss in transit (vol/sec). This is the
            total loss across all of the downstream links:
            *'surface_water__discharge_loss'*
        -  Node array containing downstream-to-upstream ordered list of node
           IDs: *'flow__upstream_node_order'*
        -  Node array of all but the first element of the delta data structure:
            *flow__data_structure_delta*. The first element is always zero.
        -  At-grid D data structure: *flow__data_structure_D*

    The FlowDirector component will add additional ModelGrid fields; see the
    `FlowAccumulator component <https://landlab.readthedocs.io/en/latest/landlab.components.flow_accum.html>`_
    for full details. These are:

        -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
            there is no receiver: *'flow__receiver_node'*
        -  Node array of flow proportions: *'flow__receiver_proportions'*
        -  Node array of links carrying flow:  *'flow__link_to_receiver_node'*
        -  Node array of downhill slopes from each receiver:
            *'topographic__steepest_slope'*
        -  Boolean node array of all local lows: *'flow__sink_flag'*

    The primary method of this class is :func:`run_one_step`.

    `run_one_step` takes the optional argument update_flow_director (default is
    True) that determines if the flow_director is re-run before flow is
    accumulated.

    Parameters
    ----------
    grid : ModelGrid
        A Landlab grid.
    surface : field name at node or array of length node
        The surface to direct flow across.
    flow_director : string, class, instance of class.
        A string of method or class name (e.g. 'D8' or 'FlowDirectorD8'), an
        uninstantiated FlowDirector class, or an instance of a FlowDirector
        class. This sets the method used to calculate flow directions.
        Default is 'FlowDirectorSteepest'
    runoff_rate : field name, array, or float, optional (m/time)
        If provided, sets the runoff rate and will be assigned to the grid
        field 'water__unit_flux_in'. If a spatially and and temporally variable
        runoff rate is desired, pass this field name and update the field
        through model run time. If both the field and argument are present at
        the time of initialization, runoff_rate will *overwrite* the field. If
        neither are set, defaults to spatially constant unit input.
    depression_finder : string, class, instance of class, optional
        A string of class name (e.g., 'DepressionFinderAndRouter'), an
        uninstantiated DepressionFinder class, or an instance of a
        DepressionFinder class.
        This sets the method for depression finding.
    loss_function : Python function, optional
        A function of the form f(Qw, [node_ID, [linkID, [grid]]]), where Qw is
        the discharge at a node, node_ID the ID of the node at which the loss
        is to be calculated, linkID is the ID of the link down which the
        outflow drains (or a d8 ID if the routing is d8), and grid is a Landlab
        ModelGrid. The function then returns the new discharge at the node
        after the function is applied.
        Note that if a linkID is needed, a nodeID must also be specified, even
        if only as a dummy parameter; similarly, if a grid is to be passed, all
        of the preceding parameters must be specified. Both nodeID and linkID
        are required to permit spatially variable losses, and also losses
        dependent on flow path geometry (e.g., flow length). The grid is passed
        to allow fields or grid properties describing values across the grid
        to be accessed for the loss calculation (see examples).
        This function expects (float, [int, [int, [ModelGrid]]]), and
        return a single float, the new discharge value. This behavior is
        verified during component instantiation.
    **kwargs : optional
        Any additional parameters to pass to a FlowDirector or
        DepressionFinderAndRouter instance (e.g., partion_method for
        FlowDirectorMFD). This will have no effect if an instantiated
        component is passed using the flow_director or depression_finder
        keywords.

    Examples
    --------
    These examples pertain only to the LossyFlowAccumulator. See the main
    FlowAccumulator documentation for more generic and comprehensive examples.

    First, a very simple example. Here's a 50% loss of discharge every time
    flow moves along a node:

    >>> import numpy as np
    >>> from landlab import RasterModelGrid, HexModelGrid
    >>> from landlab.components import FlowDirectorSteepest
    >>> from landlab.components import DepressionFinderAndRouter

    >>> mg = RasterModelGrid((3, 5), xy_spacing=(2, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    >>> z = mg.add_field('topographic__elevation',
    ...                  mg.node_x + mg.node_y,
    ...                  at='node')

    >>> def mylossfunction(qw):
    ...     return 0.5 * qw

    >>> fa = LossyFlowAccumulator(mg, 'topographic__elevation',
    ...                           flow_director=FlowDirectorSteepest,
    ...                           routing='D4', loss_function=mylossfunction)
    >>> fa.run_one_step()

    >>> mg.at_node['drainage_area'].reshape(mg.shape)
    array([[ 0.,  0.,  0.,  0.,  0.],
           [ 6.,  6.,  4.,  2.,  0.],
           [ 0.,  0.,  0.,  0.,  0.]])
    >>> mg.at_node['surface_water__discharge'].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 1.75,  3.5 ,  3.  ,  2.  ,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])
    >>> mg.at_node['surface_water__discharge_loss'].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 0.  ,  1.75,  1.5 ,  1.  ,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])

    Here we use a spatially distributed field to derive loss terms, and also
    use a filled, non-raster grid.

    >>> dx=(2./(3.**0.5))**0.5  # area to be 100.
    >>> hmg = HexModelGrid(5,3, dx, xy_of_lower_left=(-1.0745, 0.))
    >>> z = hmg.add_field('topographic__elevation',
    ...                   hmg.node_x**2 + np.round(hmg.node_y)**2,
    ...                   at = 'node')
    >>> z[9] = -10.  # poke a hole
    >>> lossy = hmg.add_zeros('node', 'mylossterm', dtype=float)
    >>> lossy[14] = 1.  # suppress all flow from node 14

    Without loss looks like this:

    >>> fa = LossyFlowAccumulator(hmg, 'topographic__elevation',
    ...                           flow_director=FlowDirectorSteepest,
    ...                           depression_finder=DepressionFinderAndRouter)
    >>> fa.run_one_step()
    >>> hmg.at_node['flow__receiver_node'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0,  1,  2,
            3,  0,  9,  6,
            7,  9,  4,  9, 11,
           12,  9,  9, 15,
           16, 17, 18])
    >>> hmg.at_node['drainage_area'] # doctest: +NORMALIZE_WHITESPACE
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> hmg.at_node['surface_water__discharge'] # doctest: +NORMALIZE_WHITESPACE
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])

    With loss looks like this:

    >>> def mylossfunction2(Qw, nodeID, linkID, grid):
    ...     return (1. - grid.at_node['mylossterm'][nodeID]) * Qw
    >>> fa = LossyFlowAccumulator(hmg, 'topographic__elevation',
    ...                           flow_director=FlowDirectorSteepest,
    ...                           depression_finder=DepressionFinderAndRouter,
    ...                           loss_function=mylossfunction2)
    >>> fa.run_one_step()
    >>> hmg.at_node['drainage_area'] # doctest: +NORMALIZE_WHITESPACE
    array([ 7.,  0.,  0.,
            0.,  7.,  1.,  0.,
            0.,  1.,  6.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> hmg.at_node['surface_water__discharge'] # doctest: +NORMALIZE_WHITESPACE
    array([ 6.,  0.,  0.,
            0.,  6.,  1.,  0.,
            0.,  1.,  5.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])
    >>> np.allclose(hmg.at_node['surface_water__discharge_loss'],
    ...             lossy*hmg.at_node['surface_water__discharge'])
    True

    (Loss is only happening from the node, 14, that we set it to happen at.)


    Finally, note we can use the linkIDs to create flow-length-dependent
    effects:

    >>> from landlab.components import FlowDirectorMFD
    >>> mg = RasterModelGrid((4, 6), xy_spacing=(1, 2))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    >>> z = mg.add_field('node', 'topographic__elevation', 2.*mg.node_x)
    >>> z[9] = 8.
    >>> z[16] = 6.5  # force the first node sideways

    >>> L = mg.add_zeros('node', 'spatialloss')
    >>> mg.at_node['spatialloss'][9] = 1.
    >>> mg.at_node['spatialloss'][13] = 1.
    >>> def fancyloss(Qw, nodeID, linkID, grid):
    ...     # now a true transmission loss:
    ...     Lt = (1. - 1./grid.length_of_link[linkID]**2)
    ...     Lsp = grid.at_node['spatialloss'][nodeID]
    ...     return Qw * (1. - Lt) * (1. - Lsp)

    >>> fa = LossyFlowAccumulator(mg, 'topographic__elevation',
    ...                           flow_director=FlowDirectorMFD,
    ...                           loss_function=fancyloss)
    >>> fa.run_one_step()

    >>> mg.at_node['drainage_area'].reshape(mg.shape)
    array([[  0. ,   0. ,   0. ,   0. ,   0. ,   0. ],
           [  5.6,   5.6,   3.6,   2. ,   2. ,   0. ],
           [ 10.4,  10.4,   8.4,   6.4,   4. ,   0. ],
           [  0. ,   0. ,   0. ,   0. ,   0. ,   0. ]])
    >>> mg.at_node['surface_water__discharge'].reshape(mg.shape)
    array([[ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
           [ 4. ,  4. ,  2. ,  2. ,  2. ,  0. ],
           [ 0. ,  8.5,  6.5,  4.5,  2.5,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ]])
    """

    _name = "LossyFlowAccumulator"

    _input_var_names = ("topographic__elevation", "water__unit_flux_in")

    _output_var_names = (
        "drainage_area",
        "surface_water__discharge",
        "surface_water__discharge_loss",
        "flow__upstream_node_order",
        "flow__nodes_not_in_stack",
        "flow__data_structure_delta",
        "flow__data_structure_D",
    )

    _var_units = {
        "topographic__elevation": "m",
        "flow__receiver_node": "m",
        "water__unit_flux_in": "m/s",
        "drainage_area": "m**2",
        "surface_water__discharge": "m**3/s",
        "surface_water__discharge_loss": "m**3/s",
        "flow__upstream_node_order": "-",
        "flow__data_structure_delta": "-",
        "flow__data_structure_D": "-",
        "flow__nodes_not_in_stack": "-",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "flow__receiver_node": "node",
        "water__unit_flux_in": "node",
        "drainage_area": "node",
        "surface_water__discharge": "node",
        "surface_water__discharge_loss": "node",
        "flow__upstream_node_order": "node",
        "flow__nodes_not_in_stack": "grid",
        "flow__data_structure_delta": "node",
        "flow__data_structure_D": "grid",
    }
    _var_doc = {
        "topographic__elevation": "Land surface topographic elevation",
        "flow__receiver_node": "Node array of receivers (node that "
        + "receives flow from current node)",
        "drainage_area": "Upstream accumulated surface area contributing "
        + "to the node's discharge",
        "surface_water__discharge": "Discharge of water through each node",
        "surface_water__discharge_loss": "Total volume of water per second "
        + "lost during all flow out of the node",
        "water__unit_flux_in": "External volume water per area per time "
        + "input to each node (e.g., rainfall rate)",
        "flow__upstream_node_order": "Node array containing "
        + "downstream-to-upstream ordered list of node IDs",
        "flow__data_structure_delta": "Node array containing the "
        + "elements delta[1:] of the data structure 'delta' used for "
        + "construction of the downstream-to-upstream node array",
        "flow__data_structure_D": "Grid array containing the data structure "
        + "D used for construction of the downstream-to-upstream node array",
        "flow__nodes_not_in_stack": "Boolean value indicating if there "
        + "are any nodes that have not yet been added to the stack stored "
        + "in flow__upstream_node_order.",
    }

    def __init__(
        self,
        grid,
        surface="topographic__elevation",
        flow_director="FlowDirectorSteepest",
        runoff_rate=None,
        depression_finder=None,
        loss_function=None,
        **kwargs
    ):
        """Initialize the FlowAccumulator component.

        Saves the grid, tests grid type, tests imput types and
        compatability for the flow_director and depression_finder
        keyword arguments, tests the argument of runoff_rate, and
        initializes new fields.
        """
        super(LossyFlowAccumulator, self).__init__(
            grid,
            surface=surface,
            flow_director=flow_director,
            runoff_rate=runoff_rate,
            depression_finder=depression_finder,
            **kwargs
        )

        if loss_function is not None:
            if sys.version_info[0] >= 3:
                sig = signature(loss_function)
                num_params = len(sig.parameters)
            else:  # Python 2
                num_params = loss_function.func_code.co_argcount
            # save the func for loss, and do a quick test on its inputs:
            if num_params == 1:
                # check the func takes a single value and turns it into a new
                # single value:
                if not isinstance(loss_function(1.), float):
                    raise TypeError(
                        "The loss_function should take a float, and return " "a float."
                    )
                # now, for logical consistency in our calls to
                # find_drainage_area_and_discharge, wrap the func so it has two
                # arguments:

                def lossfunc(Qw, dummyn, dummyl, dummygrid):
                    return float(loss_function(Qw))

                self._lossfunc = lossfunc

            elif num_params == 2:
                # check the func takes a single value and turns it into a new
                # single value:
                if not isinstance(loss_function(1., 0), float):
                    raise TypeError(
                        "The loss_function should take (float, int), and "
                        "return a float."
                    )
                # now, for logical consistency in our calls to
                # find_drainage_area_and_discharge, wrap the func so it has two
                # arguments:

                def lossfunc(Qw, nodeID, dummyl, dummygrid):
                    return float(loss_function(Qw, nodeID))

                self._lossfunc = lossfunc

            elif num_params == 3:
                # check the func takes (float, int) and turns it into a new
                # single value:
                if not isinstance(loss_function(1., 0, 0), float):
                    raise TypeError(
                        "The loss_function should take (float, int, int), "
                        "and return a float."
                    )

                def lossfunc(Qw, nodeID, linkID, dummygrid):
                    return float(loss_function(Qw, nodeID, linkID))

                self._lossfunc = lossfunc

            elif num_params == 4:
                # this time, the test is too hard to implement cleanly so just
                self._lossfunc = loss_function
            else:
                raise ValueError(
                    "The loss_function must have only a single argument, "
                    "which should be the discharge at a node; a pair of "
                    "arguments, which should be the discharge at a node and "
                    "the node ID; or three arguments, which should be the "
                    "discharge at a node, the node ID, and the link along "
                    "which that discharge will flow."
                )
        else:
            # make a dummy
            def lossfunc(Qw, dummyn, dummyl, dummygrid):
                return float(Qw)

            self._lossfunc = lossfunc

        # add the new loss discharge field if necessary:
        if "surface_water__discharge_loss" not in grid.at_node:
            self.grid.add_zeros(
                "node", "surface_water__discharge_loss", dtype=float, noclobber=False
            )

    def _accumulate_A_Q_to_one(self, s, r):
        """Accumulate area and discharge for a route-to-one scheme."""
        link = self._grid.at_node["flow__link_to_receiver_node"]
        a, q = flow_accum_bw.find_drainage_area_and_discharge_lossy(
            s,
            r,
            link,
            self._lossfunc,
            self._grid,
            self.node_cell_area,
            self._grid.at_node["water__unit_flux_in"],
        )
        return a, q

    def _accumulate_A_Q_to_n(self, s, r, p):
        """Accumulate area and discharge for a route-to-one scheme."""
        link = self._grid.at_node["flow__link_to_receiver_node"]
        a, q = flow_accum_to_n.find_drainage_area_and_discharge_to_n_lossy(
            s,
            r,
            link,
            p,
            self._lossfunc,
            self._grid,
            self.node_cell_area,
            self._grid.at_node["water__unit_flux_in"],
        )
        return a, q


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



#! /usr/env/python

"""
flow_director_steepest.py: provides the component FlowDirectorSteepest.

This components finds the steepest single-path steepest descent flow
directions. It is equivalent to D4 method in the special case of a raster grid
in that it does not consider diagonal links between nodes. For that capability,
use FlowDirectorD8.
"""

import numpy as np

from landlab import (
    BAD_INDEX_VALUE,
    FIXED_GRADIENT_BOUNDARY,
    FIXED_VALUE_BOUNDARY,
    VoronoiDelaunayGrid,
)
from landlab.components.flow_director import flow_direction_DN
from landlab.components.flow_director.flow_director_to_one import _FlowDirectorToOne


class FlowDirectorSteepest(_FlowDirectorToOne):

    """Single-path (steepest direction) flow direction without diagonals.

    This components finds the steepest single-path steepest descent flow
    directions. It is equivalent to D4 method in the special case of a raster
    grid in that it does not consider diagonal links between nodes. For that
    capability, use FlowDirectorD8.

    Stores as ModelGrid fields:

    -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
       there is no receiver: *'flow__receiver_node'*
    -  Node array of steepest downhill slopes:
       *'topographic__steepest_slope'*
    -  Node array containing ID of link that leads from each node to its
       receiver, or BAD_INDEX_VALUE if no link:
       *'flow__link_to_receiver_node'*
    -  Boolean node array of all local lows: *'flow__sink_flag'*
    -  Link array identifing if flow goes with (1) or against (-1) the link
       direction: *'flow_link_direction'*

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------

    This method works for both raster and irregular grids. First we will look
    at a raster example, and then an irregular example.

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowDirectorSteepest
    >>> mg = RasterModelGrid((3,3), xy_spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x + mg.node_y,
    ...     at = 'node'
    ... )
    >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])
    >>> fd.run_one_step()
    >>> mg.at_node['flow__receiver_node']
    array([0, 1, 2, 3, 1, 5, 6, 7, 8])
    >>> mg.at_node['topographic__steepest_slope']
    array([ 0.,  0.,  0.,  0.,  1.,  0.,  0.,  0.,  0.])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([-1, -1, -1, -1,  3, -1, -1, -1, -1])
    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1, 1, 0, 1, 1, 1, 1], dtype=int8)
    >>> mg_2 = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    >>> topographic__elevation = np.array([0.,  0.,  0., 0.,
    ...                                    0., 21., 10., 0.,
    ...                                    0., 31., 20., 0.,
    ...                                    0., 32., 30., 0.,
    ...                                    0.,  0.,  0., 0.])
    >>> _ = mg_2.add_field(
    ...     'node',
    ...     'topographic__elevation',
    ...     topographic__elevation
    ... )
    >>> mg_2.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> fd_2 = FlowDirectorSteepest(mg_2)
    >>> fd_2.run_one_step()
    >>> mg_2.at_node['flow__receiver_node'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0,  1,  2,  3,
            4,  1,  2,  7,
            8, 10,  6, 11,
           12, 14, 10, 15,
           16, 17, 18, 19])

    And the at-link field ``'flow_link_direction'`` indicates if the flow along
    the link is with or against the direction indicated by ``'link_dirs_at_node'``
    (from tail node to head node).

    >>> mg_2.at_link['flow_link_direction']
    array([ 0,  0,  0,  0, -1, -1,  0,  0,  0,  0,  0,  0, -1,  0,  0,  1,  0,
        0,  0, -1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0], dtype=int8)

    This indicates that flow on links 4, 5, 12, and 19 goes against the
    topologic ordering -- that is that flow goes from head node to tail node --
    and that flow goes with the topologic ordering on links 15 and 22. All other
    links have no flow on them.

    The FlowDirectorSteepest attribute ``flow_link_direction_at_node`` indicates
    the link flow direction (with or against topology directions) for all links
    at node. The ordering of links at node mirrors the grid attribute
    ``links_at_node``.

    >>> fd_2.flow_link_direction_at_node()
    array([[ 0,  0,  0,  0],
           [ 0, -1,  0,  0],
           [ 0, -1,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0, -1],
           [ 0, -1,  0, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 1,  0,  0,  0],
           [ 0, -1,  1, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 1,  0,  0,  0],
           [ 0,  0,  1, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0]], dtype=int8)

    For example, this indicates that node 10 has flow going along three links
    that are attached to it. The link to the East has no flow, the link to the
    North has flow going against the topologic direction, the link to the West
    has flow going with the topologic direction, and the link to the South has
    flow going against the topologic direction.

    In many use cases, one might want to know which links are bringing flow into
    or out of the node. The flow director attribute ``flow_link_incoming_at_node``
    provides this information. Here -1 means that flow is outgoing from the node
    and 1 means it is incoming.

    >>> fd_2.flow_link_incoming_at_node()
    array([[ 0,  0,  0,  0],
           [ 0,  1,  0,  0],
           [ 0,  1,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0, -1],
           [ 0,  1,  0, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [-1,  0,  0,  0],
           [ 0,  1,  1, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [-1,  0,  0,  0],
           [ 0,  0,  1, -1],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0],
           [ 0,  0,  0,  0]], dtype=int8)

    So if one wanted to identify the source nodes at node, you would do the
    following:

    >>> np.where(fd_2.flow_link_incoming_at_node() == 1, mg_2.adjacent_nodes_at_node, -1)
    array([[-1, -1, -1, -1],
           [-1,  5, -1, -1],
           [-1,  6, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, 10, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, 14,  9, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, 13, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1],
           [-1, -1, -1, -1]])

    The flow directors also have the ability to return the flow receiver nodes

    >>> receiver = fd.direct_flow()
    >>> receiver
    array([0, 1, 2,
           3, 1, 5,
           6, 7, 8])

    For the second example we will use a Hexagonal Model Grid, a special type
    of Voroni Grid that has regularly spaced hexagonal cells.

    >>> from landlab import HexModelGrid
    >>> mg = HexModelGrid(5,3)
    >>> _ = mg.add_field(
    ...     'topographic__elevation',
    ...     mg.node_x + np.round(mg.node_y),
    ...     at = 'node'
    ... )
    >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
    >>> fd.surface_values
    array([ 1. ,  2. ,  3. ,
        1.5,  2.5,  3.5,  4.5,
      2. ,  3. ,  4. ,  5. ,  6. ,
        3.5,  4.5,  5.5,  6.5,
            4. ,  5. ,  6. ])
    >>> fd.run_one_step()
    >>> mg.at_node['flow__receiver_node']
    array([ 0,  1,  2,
          3,  0,  1,  6,
        7,  3,  4,  5,  11,
          12,  8,  9, 15,
            16, 17, 18])
    >>> mg.at_node['topographic__steepest_slope']
    array([ 0. ,  0. ,  0. ,
        0. ,  1.5,  1.5,   0. ,
      0. ,  1.5,  1.5,  1.5,  0. ,
        0. ,  1.5,  1.5,  0. ,
            0. ,  0. ,  0. ])
    >>> mg.at_node['flow__link_to_receiver_node']
    array([-1, -1, -1,
         -1,  3,  5, -1,
       -1, 12, 14, 16, -1,
         -1, 25, 27, -1,
           -1, -1, -1])
    >>> mg.at_node['flow__sink_flag']
    array([1, 1, 1,
          1, 0, 0, 1,
         1, 0, 0, 0, 1,
          1, 0, 0, 1,
            1, 1, 1], dtype=int8)
    >>> receiver = fd.direct_flow()
    >>> receiver
    array([ 0,  1,  2,
          3,  0,  1,  6,
        7,  3,  4,  5, 11,
         12,  8,  9, 15,
          16, 17, 18])
    """

    _name = "FlowDirectorSteepest"

    def __init__(self, grid, surface="topographic__elevation"):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        surface : field name at node or array of length node, optional
            The surface to direct flow across, default is field at node:
            topographic__elevation,.
        """
        self.method = "D4"
        super(FlowDirectorSteepest, self).__init__(grid, surface)
        self._is_Voroni = isinstance(self._grid, VoronoiDelaunayGrid)

        # create a : 'flow_link_direction' field if it does not exist yest
        if "flow_link_direction" not in self._grid.at_link:
            self._flow_link_direction = grid.add_field(
                "flow_link_direction",
                grid.zeros(at="link", dtype=np.int8),
                at="link",
                dtype=int,
            )
        else:
            self._flow_link_direction = grid.at_link["flow_link_direction"]

        self.updated_boundary_conditions()

    def updated_boundary_conditions(self):
        """Method to update FlowDirectorSteepest when boundary conditions
        change.

        Call this if boundary conditions on the grid are updated after
        the component is instantiated.
        """
        self._active_links = self.grid.active_links
        self._activelink_tail = self.grid.node_at_link_tail[self.grid.active_links]
        self._activelink_head = self.grid.node_at_link_head[self.grid.active_links]

    def run_one_step(self):
        """Find flow directions and save to the model grid.

        run_one_step() checks for updated boundary conditions, calculates
        slopes on links, finds baselevel nodes based on the status at node,
        calculates flow directions, and saves results to the grid.

        An alternative to direct_flow() is run_one_step() which does the same
        things but also returns the receiver nodes not return values.
        """
        self.direct_flow()

    def direct_flow(self):
        """Find flow directions, save to the model grid, and return receivers.

        direct_flow() checks for updated boundary conditions, calculates
        slopes on links, finds baselevel nodes based on the status at node,
        calculates flow directions, saves results to the grid, and returns a
        at-node array  of receiver nodes. This array is stored in the grid at:
        grid['node']['flow__receiver_node']

        An alternative to direct_flow() is run_one_step() which does the same
        things but also returns a at-node array  of receiver nodes. This array
        is stored in the grid at:
        grid['node']['flow__receiver_node']
        """
        self._check_updated_bc()

        # update the surface, if it was provided as a model grid field.
        self._changed_surface()

        # step 1. Calculate link slopes at active links only.
        all_grads = -self._grid.calc_grad_at_link(self.surface_values)
        link_slope = all_grads[self._grid.active_links]

        # Step 2. Find and save base level nodes.
        (baselevel_nodes,) = np.where(
            np.logical_or(
                self._grid.status_at_node == FIXED_VALUE_BOUNDARY,
                self._grid.status_at_node == FIXED_GRADIENT_BOUNDARY,
            )
        )

        # Calculate flow directions
        receiver, steepest_slope, sink, recvr_link = flow_direction_DN.flow_directions(
            self.surface_values,
            self._active_links,
            self._activelink_tail,
            self._activelink_head,
            link_slope,
            grid=self._grid,
            baselevel_nodes=baselevel_nodes,
        )

        # Save the four ouputs of this component.
        self._grid["node"]["flow__receiver_node"][:] = receiver
        self._grid["node"]["topographic__steepest_slope"][:] = steepest_slope
        self._grid["node"]["flow__link_to_receiver_node"][:] = recvr_link
        self._grid["node"]["flow__sink_flag"][:] = np.zeros_like(receiver, dtype=bool)
        self._grid["node"]["flow__sink_flag"][sink] = True

        # determine link directions
        self._determine_link_directions()

        return receiver

    def _determine_link_directions(self):
        """Determine link directions and set flow_link_direction field.

        This routine is slightly different between the route-to-one and
        route-to-many methods.

        It works when DepressionFinderAndRouter is run.
        """
        # start by re-setting all links to zero.
        self._flow_link_direction[:] = 0

        # identify where flow is active on links
        is_active_flow_link = self.links_to_receiver != BAD_INDEX_VALUE

        # make an array that says which link ID is active
        active_flow_links = self.links_to_receiver[is_active_flow_link]

        # for each of those links, the position is the upstream node
        upstream_node_of_active_flow_link = np.where(is_active_flow_link)[0]

        # get the head node
        head_node_at_active_flow_link = self._grid.node_at_link_head[active_flow_links]

        # if head node is upstream node = -1, else 1
        self._flow_link_direction[
            active_flow_links[
                head_node_at_active_flow_link == upstream_node_of_active_flow_link
            ]
        ] = -1
        self._flow_link_direction[
            active_flow_links[
                head_node_at_active_flow_link != upstream_node_of_active_flow_link
            ]
        ] = 1

    def flow_link_direction_at_node(self):
        """Return array of flow link direction at node.

        This property mirrors links_at_node and indicates the relationship
        between the flow direction (determined based on the elevation of nodes)
        and the topologic link direction (in which the head and tail nodes are
        defined based on relative position in x-y space).

        It has the shape (number of nodes, maximum number of links at node).

        Recall that the standard landlab link direction goes from the tail node
        to the head node.

        A value of zero indicates that the link does not exist or is not
        active.

        A value of -1 indicates that water flow based on
        ``flow__link_to_receiver_node`` goes from head node to tail node, while
        a value of 1 indicates that water flow goes from tail node to head
        node.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowDirectorSteepest
        >>> mg = RasterModelGrid((3,3))
        >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> _ = mg.add_field(
        ...     'topographic__elevation',
        ...     mg.node_x + mg.node_y,
        ...     at = 'node'
        ... )
        >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
        >>> fd.run_one_step()
        >>> fd.flow_link_direction_at_node()
        array([[ 0,  0,  0,  0],
               [ 0, -1,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0]], dtype=int8)

        This method will be updated when the DepressionFinderAndRouter is run.

        First, without DepressionFinderAndRouter:

        >>> from landlab.components import FlowAccumulator
        >>> mg1 = RasterModelGrid((5, 5))
        >>> z1 = mg1.add_field(
        ...      'node',
        ...      'topographic__elevation',
        ...      mg1.x_of_node+2 * mg1.y_of_node
        ... )
        >>> z1[12] -= 5
        >>> mg1.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> fa1 = FlowAccumulator(mg1, flow_director='Steepest')
        >>> fa1.run_one_step()
        >>> fa1.flow_director.links_to_receiver
        array([-1, -1, -1, -1, -1,
               -1,  5, 15,  7, -1,
               -1, 19, -1, 20, -1,
               -1, 23, 24, 25, -1,
               -1, -1, -1, -1, -1])
        >>> fa1.flow_director.flow_link_direction_at_node()
        array([[ 0,  0,  0,  0],
               [ 0, -1,  0,  0],
               [ 0,  0,  0,  0],
               [ 0, -1,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  1,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 1, -1,  0,  0],
               [-1, -1,  1,  1],
               [ 0, -1, -1,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  0,  0, -1],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0]], dtype=int8)

        Next with DepressionFinderAndRouter:

        >>> mg2 = RasterModelGrid((5, 5))
        >>> z2 = mg2.add_field(
        ...      'node',
        ...      'topographic__elevation',
        ...      mg2.x_of_node+2 * mg2.y_of_node
        ... )
        >>> z2[12] -= 5
        >>> mg2.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> fa2 = FlowAccumulator(
        ...       mg2,
        ...       flow_director='Steepest',
        ...       depression_finder='DepressionFinderAndRouter',
        ...       routing='D4'
        ... )
        >>> fa2.run_one_step()
        >>> fa2.flow_director.links_to_receiver
        array([-1, -1, -1, -1, -1,
               -1,  5,  6,  7, -1,
               -1, 19, 15, 20, -1,
               -1, 23, 24, 25, -1,
               -1, -1, -1, -1, -1])
        >>> fa2.flow_director.flow_link_direction_at_node()
        array([[ 0,  0,  0,  0],
               [ 0, -1,  0,  0],
               [ 0, -1,  0,  0],
               [ 0, -1,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0, -1,  0, -1],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 1, -1,  0,  0],
               [-1, -1,  1, -1],
               [ 0, -1, -1,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  0,  0, -1],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0]], dtype=int8)
        """
        flow_link_direction_at_node = self.flow_link_direction[self._grid.links_at_node]
        flow_to_bad = self._grid.links_at_node == BAD_INDEX_VALUE
        flow_link_direction_at_node[flow_to_bad] = 0

        return flow_link_direction_at_node

    def flow_link_incoming_at_node(self):
        """Return array that mirrors links at node and indicates incoming flow.

        This array has the shape
        (number of nodes, maximum number of links at node).

        Incoming flow is indicated as 1 and outgoing as -1. 0 indicates
        that no flow moves along the link or that the link does not exist.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowDirectorSteepest
        >>> mg = RasterModelGrid((3,3))
        >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> _ = mg.add_field(
        ...     'topographic__elevation',
        ...     mg.node_x + mg.node_y,
        ...     at = 'node'
        ... )
        >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
        >>> fd.run_one_step()
        >>> fd.flow_link_incoming_at_node()
        array([[ 0,  0,  0,  0],
               [ 0,  1,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0, -1],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0],
               [ 0,  0,  0,  0]], dtype=int8)
        """

        incoming_at_node = (
            self.flow_link_direction_at_node() * self._grid.link_dirs_at_node
        )
        return incoming_at_node

    @property
    def flow_link_direction(self):
        """Return array of flow link direction.

        This property indicates the relationship between the flow direction
        (determined based on the elevation of nodes) and the topologic link
        direction (in which the head and tail nodes are defined based on
        relative position in x-y space).

        It has the shape (number_of_links,).

        Recall that the standard landlab link direction goes from the tail node
        to the head node.

        A value of zero indicates that the link does not exist or is not
        active.

        A value of -1 indicates that water flow based on
        ``flow__link_to_receiver_node`` goes from head node to tail node, while
        a value of 1 indicates that water flow goes from tail node to head
        node.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowDirectorSteepest
        >>> mg = RasterModelGrid((3,3))
        >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> _ = mg.add_field(
        ...     'topographic__elevation',
        ...     mg.node_x + mg.node_y,
        ...     at = 'node'
        ... )
        >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
        >>> fd.run_one_step()
        >>> fd.flow_link_direction
        array([ 0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0], dtype=int8)
        """
        return self._flow_link_direction

    def upstream_node_at_link(self):
        """At-link array of the upstream node based on flow direction.

        BAD_INDEX_VALUE is given if no upstream node is defined.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowDirectorSteepest
        >>> mg = RasterModelGrid((3,3))
        >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> _ = mg.add_field(
        ...     'topographic__elevation',
        ...     mg.node_x + mg.node_y,
        ...     at = 'node'
        ... )
        >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
        >>> fd.run_one_step()
        >>> fd.upstream_node_at_link()
        array([-1, -1, -1,  4, -1, -1, -1, -1, -1, -1, -1, -1])
        """
        out = -1 * self._grid.ones(at="link", dtype=int)
        out[self._flow_link_direction == 1] = self._grid.node_at_link_tail[
            self._flow_link_direction == 1
        ]
        out[self._flow_link_direction == -1] = self._grid.node_at_link_head[
            self._flow_link_direction == -1
        ]
        return out

    def downstream_node_at_link(self):
        """At-link array of the downstream node based on flow direction.

        BAD_INDEX_VALUE is given if no downstream node is defined.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowDirectorSteepest
        >>> mg = RasterModelGrid((3,3))
        >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
        >>> _ = mg.add_field(
        ...     'topographic__elevation',
        ...     mg.node_x + mg.node_y,
        ...     at = 'node'
        ... )
        >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
        >>> fd.run_one_step()
        >>> fd.downstream_node_at_link()
        array([-1, -1, -1,  1, -1, -1, -1, -1, -1, -1, -1, -1])
        """
        out = -1 * self._grid.ones(at="link", dtype=int)
        out[self._flow_link_direction == 1] = self._grid.node_at_link_head[
            self._flow_link_direction == 1
        ]
        out[self._flow_link_direction == -1] = self._grid.node_at_link_tail[
            self._flow_link_direction == -1
        ]
        return out


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



#! /usr/bin/env python
"""Unit tests for landlab.io.netcdf module."""

import os

import pytest

from landlab import RasterModelGrid
from landlab.io import (
    MismatchGridDataSizeError,
    MismatchGridXYLowerLeft,
    MismatchGridXYSpacing,
)
from landlab.io.netcdf import WITH_NETCDF4, read_netcdf

_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

grid_mapping_keys = [
    "grid_mapping_name",
    "longitude_of_central_meridian",
    "false_easting",
    "false_northing",
    "latitude_of_projection_origin",
    "scale_factor_at_central_meridian",
    "long_name",
    "longitude_of_prime_meridian",
    "semi_major_axis",
    "inverse_flattening",
    "spatial_ref",
    "GeoTransform",
]


@pytest.mark.skipif(not WITH_NETCDF4, reason="netCDF4 package not installed")
def test_read_netcdf4_bad_field_name():
    with pytest.raises(ValueError):
        read_netcdf(
            os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"),
            name="not_surface__elevation",
        )


def test_read_netcdf3_64bit():
    """Test read_netcdf for with 64-bit netcdf3 format."""
    grid = read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf3-64bit.nc"))
    assert grid.shape == (4, 3)


@pytest.mark.skipif(not WITH_NETCDF4, reason="netCDF4 package not installed")
def test_read_netcdf4():
    """Test read_netcdf with netcdf4 format."""
    grid = read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"))
    assert grid.shape == (4, 3)

    grid = read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"))
    assert grid.shape == (4, 3)


@pytest.mark.skipif(not WITH_NETCDF4, reason="netCDF4 package not installed")
def test_bad_data_size():
    """Test read_netcdf with netcdf4 format."""
    grid = RasterModelGrid((10, 10))
    with pytest.raises(MismatchGridDataSizeError):
        read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"), grid=grid)


@pytest.mark.skipif(not WITH_NETCDF4, reason="netCDF4 package not installed")
def test_bad_dx():
    """Test read_netcdf with netcdf4 format."""
    grid = RasterModelGrid((4, 3), xy_spacing=10)
    with pytest.raises(MismatchGridXYSpacing):
        read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"), grid=grid)


@pytest.mark.skipif(not WITH_NETCDF4, reason="netCDF4 package not installed")
def test_bad_llc():
    """Test read_netcdf with netcdf4 format."""
    grid = RasterModelGrid((4, 3), xy_of_lower_left=(-1, -2))
    with pytest.raises(MismatchGridXYLowerLeft):
        read_netcdf(os.path.join(_TEST_DATA_DIR, "test-netcdf4.nc"), grid=grid)



#! /usr/bin/env python
"""The Landlab

:Package name: TheLandlab
:Release date: 2018-09-18
:Authors: Greg Tucker, Nicole Gasparini, Erkan Istanbulluoglu, Daniel Hobley,
    Sai Nudurupati, Jordan Adams, Eric Hutton, Katherine Barnhart, Margaux
    Mouchene, Nathon Lyons
:URL: http://csdms.colorado.edu/trac/landlab
:License: MIT
"""
from __future__ import absolute_import

from numpy import set_printoptions

from ._registry import registry
from ._version import get_versions
from .core.model_component import Component
from .core.model_parameter_dictionary import (
    MissingKeyError,
    ModelParameterDictionary,
    ParameterValueError,
)
from .core.model_parameter_loader import load_params
from .field.scalar_data_fields import FieldError
from .framework.collections import Arena, NoProvidersError, Palette
from .framework.decorators import Implements, ImplementsOrRaise
from .framework.framework import Framework
from .grid import (
    ACTIVE_LINK,
    BAD_INDEX_VALUE,
    CLOSED_BOUNDARY,
    CORE_NODE,
    FIXED_GRADIENT_BOUNDARY,
    FIXED_LINK,
    FIXED_VALUE_BOUNDARY,
    INACTIVE_LINK,
    LOOPED_BOUNDARY,
    HexModelGrid,
    ModelGrid,
    RadialModelGrid,
    RasterModelGrid,
    VoronoiDelaunayGrid,
    NetworkModelGrid,
    create_and_initialize_grid,
    create_grid,
)
from .plot import (
    analyze_channel_network_and_plot,
    imshow_cell_grid,
    imshow_grid,
    imshow_grid_at_node,
    imshow_node_grid,
)

try:
    set_printoptions(legacy="1.13")
except TypeError:
    pass
finally:
    del set_printoptions

cite_as = registry.format_citations

__all__ = [
    "registry",
    "ModelParameterDictionary",
    "MissingKeyError",
    "ParameterValueError",
    "Component",
    "Palette",
    "Arena",
    "NoProvidersError",
    "Implements",
    "ImplementsOrRaise",
    "Framework",
    "FieldError",
    "LandlabTester",
    "load_params",
    "ModelGrid",
    "HexModelGrid",
    "RadialModelGrid",
    "RasterModelGrid",
    "VoronoiDelaunayGrid",
    "NetworkModelGrid",
    "BAD_INDEX_VALUE",
    "CORE_NODE",
    "FIXED_VALUE_BOUNDARY",
    "FIXED_GRADIENT_BOUNDARY",
    "LOOPED_BOUNDARY",
    "CLOSED_BOUNDARY",
    "ACTIVE_LINK",
    "FIXED_LINK",
    "INACTIVE_LINK",
    "create_and_initialize_grid",
    "create_grid",
    "imshow_grid",
    "imshow_node_grid",
    "imshow_cell_grid",
    "imshow_grid_at_node",
    "analyze_channel_network_and_plot",
]

__version__ = get_versions()["version"]
del get_versions



#! /usr/bin/env python
"""Create landlab model grids."""
import inspect
from warnings import warn

from landlab.core import load_params, model_parameter_dictionary as mpd
from landlab.io import read_esri_ascii
from landlab.io.netcdf import read_netcdf
from landlab.values import constant, plane, random, sine

from .hex import HexModelGrid, from_dict as hex_from_dict
from .network import NetworkModelGrid
from .radial import RadialModelGrid
from .raster import RasterModelGrid, from_dict as raster_from_dict
from .voronoi import VoronoiDelaunayGrid

_MODEL_GRIDS = {
    "RasterModelGrid": RasterModelGrid,
    "HexModelGrid": HexModelGrid,
    "VoronoiDelaunayGrid": VoronoiDelaunayGrid,
    "NetworkModelGrid": NetworkModelGrid,
    "RadialModelGrid": RadialModelGrid,
}

_SYNTHETIC_FIELD_CONSTRUCTORS = {
    "plane": plane,
    "random": random,
    "sine": sine,
    "constant": constant,
}


class Error(Exception):

    """Base class for exceptions from this module."""

    pass


class BadGridTypeError(Error):

    """Raise this error for a bad grid type."""

    def __init__(self, grid_type):
        self._type = str(grid_type)

    def __str__(self):
        return self._type


_GRID_READERS = {"raster": raster_from_dict, "hex": hex_from_dict}


def create_and_initialize_grid(input_source):
    """Create and initialize a grid from a file.

    Creates, initializes, and returns a new grid object using parameters
    specified in *input_source*. *input_source* is either a
    ModelParameterDictionary instance (or, really, just dict-like) or a
    named input file.

    Parameters
    ----------
    input_source : str or dict
        Input file or ``dict`` of parameter values.

    Raises
    ------
    KeyError
        If missing a key from the input file.

    Examples
    --------
    >>> from six import StringIO
    >>> import pytest
    >>> test_file = StringIO('''
    ... GRID_TYPE:
    ... raster
    ... NUM_ROWS:
    ... 4
    ... NUM_COLS:
    ... 5
    ... GRID_SPACING:
    ... 2.5
    ... ''')
    >>> from landlab import create_and_initialize_grid
    >>> with pytest.deprecated_call():
    ...    grid = create_and_initialize_grid(test_file)
    >>> grid.number_of_nodes
    20
    """
    msg = (
        "create_and_initialize_grid is deprecated and will be removed "
        "in landlab 2.0. Use create_grid instead."
    )
    warn(msg, DeprecationWarning)
    if isinstance(input_source, dict):
        param_dict = input_source
    else:
        param_dict = mpd.ModelParameterDictionary(from_file=input_source)

    grid_type = param_dict["GRID_TYPE"]

    grid_type.strip().lower()

    # Read parameters appropriate to that type, create it, and initialize it
    try:
        grid_reader = _GRID_READERS[grid_type]
    except KeyError:
        raise BadGridTypeError(grid_type)

    # Return the created and initialized grid
    return grid_reader(param_dict)


def create_grid(file_like):
    """Create grid, initialize fields, and set boundary conditions.

    **create_grid** expects a dictionary with three keys: "grid", "fields", and
    "boundary_conditions".

    **Dictionary Section "grid"**

    The value associated with the "grid" key should itself be a dictionary
    containing the name of a Landlab model grid type as its only key. The
    following grid types are valid:

        -  :py:class:`~landlab.grid.raster.RasterModelGrid`
        -  :py:class:`~landlab.grid.voronoi.VoronoiDelaunayGrid`
        -  :py:class:`~landlab.grid.hex.HexModelGrid`
        -  :py:class:`~landlab.grid.radial.RadialModelGrid`
        -  :py:class:`~landlab.grid.network.NetworkModelGrid`

    The value associated with the grid name key is a list containing the
    arguments. If any keyword arguments are passed, they should be passed as
    the last element of the list. For example the following code block is a
    yaml file indicating a RasterModelGrid with shape (4, 5) and xy-spacing of
    (3, 4).

    .. code-block:: yaml

        grid:
          RasterModelGrid:
            - [4, 5]
            - xy_spacing: [3, 4]

    These arguments and keyword arguments will be passed to the ``__init__``
    constructor of the specified model grid. Refer to the documentation for
    each grid to determine its requirements.

    **Dictionary Section "fields"**

    Fields can be created by reading from files or by creating synthetic
    values.

    The value associated with the "fields" key is a nested set of dictionaries
    indicating where the fields are created, what the field names are, and how
    to create the fields. At the highest hierachical level, the value
    associated with the "fields" key must be a dictionary with keys indicating
    at which grid elements fields should be created (e.g. to create fields at
    node, use "at_node").

    The value associated with each "at_xxx" value is itself a dictionary
    indicating the name of the field an how it should be created. A field can
    either be created by reading from a file or creating synthetic values. The
    :py:func:`~landlab.io.netcdf.read.read_netcdf` and
    :py:func:`~landlab.io.esri_ascii.read_esri_ascii` functions, and the
    :py:mod:`synthetic fields <landlab.values.synthetic>`
    package are currently supported methods to create fields. These may be
    chained together (as is shown in the Example section below). If these
    functions do not meet your needs, we welcome contributions that extend the
    capabilities of this function.

    The following example would uses the
    :py:func:`~landlab.values.synthetic.plane` function from the synthetic
    values package to create an at_node value for the field
    topographic__elevation. The plane function adds values to a Landlab model
    grid field that lie on a plane specified by a point and a normal vector. In
    the below example the plane goes through the point (1.0, 1.0, 1.0) and has
    a normal of (-2.0, -1.0, 1.0).

    .. code-block:: yaml

        fields:
          at_node:
            topographic__elevation:
              plane:
                - point: [1, 1, 1]
                  normal: [-2, -1, 1]

    **Dictionary Section "boundary_conditions"**

    The final portion of the input dictionary calls bound functions of the
    model grid to set boundary conditions. Any valid bound function can be
    called. The specified functions are provided in a list, and called in
    order. If required, multiple functions may be called.

    Each entry to the list is a dictionary with a single key, the name of the
    bound function. The value associated with that key is a list of arguments
    and keyword arguments, similar in structure to those described above.

    For example, the following sets closed boundaries at all sides of the grid.

    .. code-block:: yaml

        boundary_conditions:
          - set_closed_boundaries_at_grid_edges:
            - True
            - True
            - True
            - True

    Parameters
    ----------
    file_like : file_like or str
        Dictionary, contents of a dictionary as a string, a file-like object,
        or the path to a file containing a YAML dictionary.

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(42)
    >>> from landlab import create_grid
    >>> p = {'grid': {'RasterModelGrid': [(4,5),
    ...                                   {'xy_spacing': (3, 4)}]
    ...               },
    ...      'fields': {'at_node': {'spam': {'plane': [{'point': (1, 1, 1),
    ...                                                'normal': (-2, -1, 1)}],
    ...                                      'random': [{'distribution': 'uniform',
    ...                                                 'low': 1,
    ...                                                 'high': 4}]
    ...                                      },
    ...                             },
    ...                 'at_link': {'eggs': {'constant': [{'where': 'ACTIVE_LINK',
    ...                                                    'constant': 12}],
    ...                                       },
    ...                             },
    ...                 },
    ...      'boundary_conditions': [
    ...                     {'set_closed_boundaries_at_grid_edges':
    ...                                         [True, True, True, True]
    ...                      }]
    ...      }
    >>> mg = create_grid(p)
    >>> mg.number_of_nodes
    20
    >>> "spam" in mg.at_node
    True
    >>> "eggs" in mg.at_link
    True
    >>> mg.x_of_node
    array([  0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.,
             0.,   3.,   6.,   9.,  12.])
    >>> mg.status_at_node
    array([4, 4, 4, 4, 4,
           4, 0, 0, 0, 4,
           4, 0, 0, 0, 4,
           4, 4, 4, 4, 4], dtype=uint8)
    >>> np.round(mg.at_node['spam'].reshape(mg.shape), decimals=2)
    array([[  0.12,   7.85,  13.2 ,  18.8 ,  23.47],
           [  3.47,   9.17,  17.6 ,  22.8 ,  29.12],
           [  7.06,  15.91,  21.5 ,  25.64,  31.55],
           [ 11.55,  17.91,  24.57,  30.3 ,  35.87]])
    """
    # part 0, parse input
    if isinstance(file_like, dict):
        dict_like = file_like
    else:
        dict_like = load_params(file_like)

    # part 1 create grid
    grid_dict = dict_like.pop("grid", None)
    if grid_dict is None:
        msg = "create_grid: no grid dictionary provided. This is required."
        raise ValueError(msg)

    for grid_type in grid_dict:
        if grid_type in _MODEL_GRIDS:
            grid_class = _MODEL_GRIDS[grid_type]
        else:
            msg = "create_grid: provided grid type not supported."
            raise ValueError

    if len(grid_dict) != 1:
        msg = (
            "create_grid: two entries to grid dictionary provided. "
            "This is not supported."
        )
        raise ValueError

    args, kwargs = _parse_args_kwargs(grid_dict.pop(grid_type))
    grid = grid_class(*args, **kwargs)

    # part two, create fields
    fields_dict = dict_like.pop("fields", {})

    # for each grid element:
    for at_group in fields_dict:
        at = at_group[3:]
        if at not in grid.groups:
            msg = (
                "create_grid: No field location ",
                "{at} ".format(at=at),
                "exists for grid types",
                "{grid}. ".format(grid=grid_type),
            )
            raise ValueError(msg)

        at_dict = fields_dict[at_group]
        # for field at grid element
        for name in at_dict:
            name_dict = at_dict[name]

            # for each function, add values.
            for func in name_dict:
                args, kwargs = _parse_args_kwargs(name_dict[func])
                if func in _SYNTHETIC_FIELD_CONSTRUCTORS:
                    # if any args, raise an error, there shouldn't be any.
                    synth_function = _SYNTHETIC_FIELD_CONSTRUCTORS[func]
                    synth_function(grid, name, at=at, **kwargs)
                elif func == "read_esri_ascii":
                    read_esri_ascii(*args, grid=grid, name=name, **kwargs)
                elif func == "read_netcdf":
                    read_netcdf(*args, grid=grid, name=name, **kwargs)
                else:
                    msg = (
                        "create_grid: Bad function ",
                        "{func} ".format(func=func),
                        "for creating a field.",
                    )
                    raise ValueError(msg)

    # part three, set boundary conditions
    bc_list = dict_like.pop("boundary_conditions", [])
    for bc_function_dict in bc_list:
        if len(bc_function_dict) != 1:
            msg = (
                "create_grid: two entries to a boundary condition function "
                "dictionary were provided. This is not supported."
            )
            raise ValueError(msg)
        for bc_function in bc_function_dict:
            args, kwargs = _parse_args_kwargs(bc_function_dict[bc_function])
            methods = dict(inspect.getmembers(grid, inspect.ismethod))
            if bc_function in methods:
                methods[bc_function](*args, **kwargs)
            else:
                msg = (
                    "create_grid: No function ",
                    "{func} ".format(func=bc_function),
                    "exists for grid types ",
                    "{grid}. ".format(grid=grid_type),
                    "If you think this type of grid should have such a ",
                    "function. Please create a GitHub Issue to discuss ",
                    "contributing it to the Landlab codebase.",
                )
                raise ValueError(msg)

    return grid


def _parse_args_kwargs(list_of_args_kwargs):
    if isinstance(list_of_args_kwargs[-1], dict):
        kwargs = list_of_args_kwargs.pop()
    else:
        kwargs = {}
    return list_of_args_kwargs, kwargs



#!/usr/env/python

import heapq
import itertools
from collections import deque

import numpy as np


class StablePriorityQueue:
    """
    Implements a stable priority queue, that tracks insertion order; i.e., this
    is used to break ties.

    See https://docs.python.org/2/library/heapq.html#priority-queue-implementation-notes
    & https://www.sciencedirect.com/science/article/pii/S0098300413001337

    Examples
    --------
    >>> q = StablePriorityQueue()
    >>> q.add_task('b', priority=2)
    >>> q.add_task('a', priority=1)
    >>> q.add_task(0, priority=0)
    >>> q.add_task('c', priority=2)
    >>> q.remove_task(0)
    >>> q.pop_task()
    'a'
    >>> q.peek_at_task()
    'b'
    >>> np.all(q.tasks_currently_in_queue() == np.array(['b', 'c']))
    True
    >>> q.pop_task()
    'b'
    >>> np.all(q.tasks_ever_in_queue() == np.array(['b', 'a', '0', 'c']))
    True

    If only ints or floats are loaded into the array, the _in_queue methods
    return arrays with the corresponding data types:

    >>> q = StablePriorityQueue()
    >>> q.add_task(2, priority=2)
    >>> q.add_task(1, priority=1)
    >>> np.issubdtype(q.tasks_currently_in_queue().dtype, np.integer)
    True
    >>> q = StablePriorityQueue()
    >>> q.add_task(np.pi)
    >>> np.issubdtype(q.tasks_currently_in_queue().dtype, np.floating)
    True

    Popping from (or peeking at) an empty queue will throw a KeyError:

    >>> q = StablePriorityQueue()
    >>> try:
    ...     q.pop_task()
    ... except KeyError:
    ...     print('No tasks left')
    No tasks left
    """

    def __init__(self):
        self._pq = []  # list of entries as a heap
        self._entry_finder = {}  # mapping of tasks to entries
        self._REMOVED = float("inf")  # placeholder for a removed task
        self._counter = itertools.count()  # unique sequence count
        self._tasks_ever_in_queue = deque([])
        # last one tracks all nodes that have ever been added

    def add_task(self, task, priority=0):
        "Add a new task or update the priority of an existing task."
        if task == self._REMOVED:
            raise ValueError("StablePriorityQueue cannot accept tasks equal to INF!")
        if task in self._entry_finder:
            self.remove_task(task)
        count = next(self._counter)
        entry = [priority, count, task]
        self._entry_finder[task] = entry
        heapq.heappush(self._pq, entry)
        self._tasks_ever_in_queue.append(task)

    def remove_task(self, task):
        "Mark an existing task as _REMOVED.  Raise KeyError if not found."
        entry = self._entry_finder.pop(task)
        entry[-1] = self._REMOVED

    def pop_task(self):
        "Remove and return the lowest priority task. Raise KeyError if empty."
        while self._pq:
            priority, count, task = heapq.heappop(self._pq)
            if task is not self._REMOVED:
                del self._entry_finder[task]
                return task
        raise KeyError("pop from an empty priority queue")

    def peek_at_task(self):
        """
        Return the lowest priority task without removal. Raise KeyError if
        empty.
        """
        while self._pq:
            priority, count, task = self._pq[0]
            if task is not self._REMOVED:
                return task
        raise KeyError("peeked at an empty priority queue")

    def tasks_currently_in_queue(self):
        "Return array of nodes currently in the queue."
        mynodes = [
            task for (priority, count, task) in self._pq if task is not self._REMOVED
        ]
        return np.array(mynodes)

    def tasks_ever_in_queue(self):
        """
        Return array of all nodes ever added to this queue object. Repeats
        are permitted.
        """
        return np.array(self._tasks_ever_in_queue)



import os

from .read import read_netcdf
from .write import write_netcdf
from .write import write_raster_netcdf

from .errors import NotRasterGridError

try:
    import netCDF4  # noqa: F401
except ImportError:
    import warnings

    warnings.warn("Unable to import netCDF4.", ImportWarning)
    WITH_NETCDF4 = False
else:
    WITH_NETCDF4 = True

NETCDF4_EXAMPLE_FILE = os.path.join(
    os.path.dirname(__file__), "tests", "data", "test-netcdf4.nc"
)
NETCDF3_64BIT_EXAMPLE_FILE = os.path.join(
    os.path.dirname(__file__), "tests", "data", "test-netcdf3-64bit.nc"
)

__all__ = [
    "read_netcdf",
    "write_netcdf",
    "write_raster_netcdf",
    "NotRasterGridError",
    "WITH_NETCDF4",
    "NETCDF4_EXAMPLE_FILE",
    "NETCDF3_64BIT_EXAMPLE_FILE",
]



from .stream_power import StreamPowerEroder
from .fastscape_stream_power import FastscapeEroder
from .stream_power_smooth_threshold import StreamPowerSmoothThresholdEroder
from .sed_flux_dep_incision import SedDepEroder


__all__ = [
    "StreamPowerEroder",
    "FastscapeEroder",
    "SedDepEroder",
    "StreamPowerSmoothThresholdEroder",
]



import pytest

from landlab import RasterModelGrid


def test_default_names():
    rmg = RasterModelGrid((4, 5))
    assert rmg.axis_name == ("y", "x")


def test_name_keyword():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=["longitude", "latitude"],
        axis_units=["degrees_east", "degrees_north"],
    )
    assert rmg.axis_name == ("longitude", "latitude")


def test_name_setter():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=["longitude", "latitude"],
        axis_units=["degrees_east", "degrees_north"],
    )
    rmg.axis_name = ("yyy", "xxx")
    assert rmg.axis_name == ("yyy", "xxx")


def test_name_setter_too_few_names():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_name = ("z",)


def test_name_setter_too_many_names():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_name = ("z", "y", "x")


def test_default_units():
    rmg = RasterModelGrid((4, 5))
    assert rmg.axis_units == ("-", "-")


def test_axis_units_keyword():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=["longitude", "latitude"],
        axis_units=["degrees_east", "degrees_north"],
    )
    assert rmg.axis_units == ("degrees_east", "degrees_north")


def test_axis_units_setter():
    rmg = RasterModelGrid(
        (4, 5),
        axis_name=["longitude", "latitude"],
        axis_units=["degrees_east", "degrees_north"],
    )
    rmg.axis_units = ("mm", "cm")
    assert rmg.axis_units == ("mm", "cm")


def test_name_setter_too_few_units():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_units = ("m",)


def test_name_setter_too_many_units():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.axis_units = ("m", "cm", "mm")




import numpy as np

from landlab import Component

from ...utils.decorators import use_file_name_or_kwds

_VALID_METHODS = set(["Grid"])


def _assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError("%s: Invalid method name" % method)


class Radiation(Component):

    """Compute 1D and 2D total incident shortwave radiation.

    Landlab component that computes 1D and 2D total incident shortwave
    radiation. This code also computes relative incidence shortwave radiation
    compared to a flat surface. Ref: Bras, Rafael L. Hydrology: an
    introduction to hydrologic science. Addison Wesley Publishing Company,
    1990.

    .. codeauthor:: Sai Nudurupati & Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import Radiation
    >>> import numpy as np

    >>> grid = RasterModelGrid((5, 4), xy_spacing=(0.2, 0.2))
    >>> rad = Radiation(grid)
    >>> rad.name
    'Radiation'
    >>> rad.input_var_names
    ('topographic__elevation',)
    >>> sorted(rad.output_var_names) # doctest: +NORMALIZE_WHITESPACE
    ['radiation__incoming_shortwave_flux',
     'radiation__net_shortwave_flux',
     'radiation__ratio_to_flat_surface']
    >>> sorted(rad.units) # doctest: +NORMALIZE_WHITESPACE
    [('radiation__incoming_shortwave_flux', 'W/m^2'),
     ('radiation__net_shortwave_flux', 'W/m^2'),
     ('radiation__ratio_to_flat_surface', 'None'),
     ('topographic__elevation', 'm')]

    >>> rad.grid.number_of_node_rows
    5
    >>> rad.grid.number_of_node_columns
    4
    >>> rad.grid is grid
    True
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    True
    >>> np.all(grid.at_node['topographic__elevation'] == 0.)
    True

    >>> grid['node']['topographic__elevation'] = np.array([
    ...       0., 0., 0., 0.,
    ...       1., 1., 1., 1.,
    ...       2., 2., 2., 2.,
    ...       3., 4., 4., 3.,
    ...       4., 4., 4., 4.])
    >>> current_time = 0.5
    >>> rad.update(current_time)
    >>> np.all(grid.at_cell['radiation__ratio_to_flat_surface'] == 0.)
    False
    """

    _name = "Radiation"

    _input_var_names = ("topographic__elevation",)

    _output_var_names = (
        "radiation__incoming_shortwave_flux",
        "radiation__ratio_to_flat_surface",
        "radiation__net_shortwave_flux",
    )

    _var_units = {
        "topographic__elevation": "m",
        "radiation__incoming_shortwave_flux": "W/m^2",
        "radiation__ratio_to_flat_surface": "None",
        "radiation__net_shortwave_flux": "W/m^2",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "radiation__incoming_shortwave_flux": "cell",
        "radiation__ratio_to_flat_surface": "cell",
        "radiation__net_shortwave_flux": "cell",
    }

    _var_doc = {
        "topographic__elevation": "elevation of the ground surface relative to some datum",
        "radiation__incoming_shortwave_flux": "total incident shortwave radiation over the time step",
        "radiation__ratio_to_flat_surface": "ratio of total incident shortwave radiation on sloped surface \
             to flat surface",
        "radiation__net_shortwave_flux": "net incident shortwave radiation over the time step",
    }

    @use_file_name_or_kwds
    def __init__(
        self,
        grid,
        method="Grid",
        cloudiness=0.2,
        latitude=34.,
        albedo=0.2,
        solar_constant=1366.67,
        clearsky_turbidity=2.,
        opt_airmass=0.,
        **kwds
    ):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        method: {'Grid'}, optional
            Currently, only default is available.
        cloudiness: float, optional
            Cloudiness.
        latitude: float, optional
            Latitude (radians).
        albedo: float, optional
            Albedo.
        solar_constant: float, optional
            Solar Constant (W/m^2).
        clearsky_turbidity: float, optional
            Clear sky turbidity.
        opt_airmass: float, optional
            Optical air mass.
        """

        self._method = method
        self._N = cloudiness
        self._latitude = latitude
        self._A = albedo
        self._Io = solar_constant
        self._n = clearsky_turbidity
        self._m = opt_airmass

        _assert_method_is_valid(self._method)

        super(Radiation, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_node:
                self.grid.add_zeros(name, at="node", units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros(name, at="cell", units=self._var_units[name])

        if "Slope" not in self.grid.at_cell:
            self.grid.add_zeros("Slope", at="cell", units="radians")

        if "Aspect" not in self.grid.at_cell:
            self.grid.add_zeros("Aspect", at="cell", units="radians")

        self._nodal_values = self.grid["node"]
        self._cell_values = self.grid["cell"]
        self._slope, self._aspect = grid.calculate_slope_aspect_at_nodes_burrough(
            vals="topographic__elevation"
        )
        #        self._slope = grid.calc_slope_of_node( \
        #                                elevs = 'topographic__elevation')
        #        self._aspect =
        self._cell_values["Slope"] = self._slope
        self._cell_values["Aspect"] = self._aspect

    def update(self, current_time, hour=12., **kwds):
        """Update fields with current loading conditions.

        Parameters
        ----------
        current_time: float
              Current time (years).
        hour: float, optional
              Hour of the day.
        """
        self._t = hour
        self._radf = self._cell_values["radiation__ratio_to_flat_surface"]
        self._Rs = self._cell_values["radiation__incoming_shortwave_flux"]
        self._Rnet = self._cell_values["radiation__net_shortwave_flux"]

        self._julian = np.floor(
            (current_time - np.floor(current_time)) * 365.25
        )  # Julian day

        self._phi = np.radians(self._latitude)  # Latitude in Radians

        self._delta = 23.45 * np.radians(
            np.cos(2 * np.pi / 365 * (172 - self._julian))
        )  # Declination angle

        self._tau = (self._t + 12.0) * np.pi / 12.0  # Hour angle

        self._alpha = np.arcsin(
            np.sin(self._delta) * np.sin(self._phi)
            + np.cos(self._delta) * np.cos(self._phi) * np.cos(self._tau)
        )  # Solar Altitude

        if self._alpha <= 0.25 * np.pi / 180.0:  # If altitude is -ve,
            self._alpha = 0.25 * np.pi / 180.0  # sun is beyond the horizon

        # Counting for Albedo, Cloudiness and Atmospheric turbidity
        self._Rgl = self._Io * np.exp(
            (-1)
            * self._n
            * (0.128 - 0.054 * np.log10(1. / np.sin(self._alpha)))
            * (1. / np.sin(self._alpha))
        )

        self._phisun = np.arctan(
            -np.sin(self._tau)
            / (
                np.tan(self._delta) * np.cos(self._phi)
                - np.sin(self._phi) * np.cos(self._tau)
            )
        )  # Sun's Azhimuth
        if self._phisun >= 0 and -np.sin(self._tau) <= 0:
            self._phisun = self._phisun + np.pi

        elif self._phisun <= 0 and -np.sin(self._tau) >= 0:
            self._phisun = self._phisun + np.pi

        self._flat = np.cos(np.arctan(0)) * np.sin(self._alpha) + np.sin(
            np.arctan(0)
        ) * np.cos(self._alpha) * np.cos(
            self._phisun - 0
        )  # flat surface reference

        # flat surface total incoming shortwave radiation
        self._Rsflat = self._Rgl * self._flat

        # flat surface Net incoming shortwave radiation
        self._Rnetflat = (1 - self._A) * (1 - 0.65 * (self._N ** 2)) * self._Rsflat

        self._sloped = np.cos(self._slope) * np.sin(self._alpha) + np.sin(
            self._slope
        ) * np.cos(self._alpha) * np.cos(self._phisun - self._aspect)
        # Ratio of cosine of solar incidence angle of sloped surface to that
        # of a flat surface
        self._radf = self._sloped / self._flat

        self._radf[self._radf <= 0.] = 0.
        self._radf[self._radf > 6.] = 6.

        # Sloped surface Toatl Incoming Shortwave Radn
        self._Rs = self._Rsflat * self._radf
        self._Rnet = self._Rnetflat * self._radf

        self._cell_values["radiation__ratio_to_flat_surface"] = self._radf
        self._cell_values["radiation__incoming_shortwave_flux"] = self._Rs
        self._cell_values["radiation__net_shortwave_flux"] = self._Rnet



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import CLOSED_BOUNDARY, CORE_NODE
from landlab.field import GroupError
from landlab.values import constant, plane, random
from landlab.values.synthetic import _plane_function, _where_to_add_values

_NORMAL = (1, 1, 1)
_POINT = (0, 0, 0)


def test_bad_grid_element_name(four_by_four_raster):
    with pytest.raises(KeyError):
        constant(four_by_four_raster, "some_flux", "not_a_grid_element", constant=1.0)


def test_bad_distribution_name(four_by_four_raster):
    with pytest.raises(ValueError):
        random(four_by_four_raster, "values", "node", distribution="not_a_distribution")


def test_vertical_plane(four_by_four_raster):
    with pytest.raises(ValueError):
        plane(four_by_four_raster, "values", normal=(0., 1., 0.))


def test_no_xy_values(four_by_four_raster):
    with pytest.raises(ValueError):
        plane(four_by_four_raster, "values", "patch")


def test_xy_node_raster(four_by_four_raster):
    vals = plane(four_by_four_raster, "values", "node", normal=_NORMAL)
    truth = _plane_function(
        four_by_four_raster.x_of_node,
        four_by_four_raster.y_of_node,
        point=_POINT,
        normal=_NORMAL,
    )
    assert_array_equal(vals, truth)


def test_xy_cell_raster(four_by_four_raster):
    vals = plane(four_by_four_raster, "values", "cell", normal=_NORMAL)
    truth = _plane_function(
        four_by_four_raster.x_of_cell,
        four_by_four_raster.y_of_cell,
        point=_POINT,
        normal=_NORMAL,
    )
    assert_array_equal(vals, truth)


def test_xy_link_raster(four_by_four_raster):
    vals = plane(four_by_four_raster, "values", "link", normal=_NORMAL)
    truth = _plane_function(
        four_by_four_raster.x_of_link,
        four_by_four_raster.y_of_link,
        point=_POINT,
        normal=_NORMAL,
    )
    assert_array_equal(vals, truth)


def test_xy_face_raster(four_by_four_raster):
    vals = plane(four_by_four_raster, "values", "face", normal=_NORMAL)
    truth = _plane_function(
        four_by_four_raster.x_of_face,
        four_by_four_raster.y_of_face,
        point=_POINT,
        normal=_NORMAL,
    )
    assert_array_equal(vals, truth)


def test_xy_node_network(simple_network):
    vals = plane(simple_network, "values", "node", normal=_NORMAL)
    truth = _plane_function(
        simple_network.x_of_node, simple_network.y_of_node, point=_POINT, normal=_NORMAL
    )
    assert_array_equal(vals, truth)


def test_xy_face_network(simple_network):
    with pytest.raises(ValueError):
        plane(simple_network, "values", "face", normal=_NORMAL)


def test_where_status_with_patches(four_by_four_raster):
    with pytest.raises(AttributeError):
        constant(four_by_four_raster, "values", "patch", where=CORE_NODE)


def test_multiple_status_node(four_by_four_raster):
    four_by_four_raster.set_closed_boundaries_at_grid_edges(True, True, False, False)
    vals = constant(
        four_by_four_raster,
        "values",
        "node",
        where=[CORE_NODE, CLOSED_BOUNDARY],
        constant=10.,
    )
    true_array = np.array(
        [0., 0., 0., 0., 0., 10., 10., 10., 0., 10., 10., 10., 10., 10., 10., 10.]
    )
    assert_array_equal(vals, true_array)


def test_where_to_add_values_where_is_none(four_by_four_raster, at):
    where = _where_to_add_values(four_by_four_raster, at, None)
    assert np.all(where == np.full(four_by_four_raster.size(at), True))


def test_where_to_add_values_where_is_everywhere(four_by_four_raster, at):
    where = np.full(four_by_four_raster.size(at), True)
    assert np.all(_where_to_add_values(four_by_four_raster, at, where) == where)


def test_where_to_add_values_where_is_nowhere(four_by_four_raster, at):
    where = np.full(four_by_four_raster.size(at), False)
    assert np.all(_where_to_add_values(four_by_four_raster, at, where) == where)


def test_where_to_add_values_where_is_somewhere(four_by_four_raster, at):
    where = np.random.randint(0, 2, size=four_by_four_raster.size(at), dtype=bool)
    assert np.all(_where_to_add_values(four_by_four_raster, at, where) == where)


def test_where_to_add_values_node_bc(four_by_four_raster, at, node_bc):
    if at == "node":
        where = _where_to_add_values(four_by_four_raster, at, node_bc)
        assert len(where) == four_by_four_raster.number_of_nodes
    elif at == "link":
        with pytest.raises(ValueError):
            _where_to_add_values(four_by_four_raster, at, node_bc)
    else:
        with pytest.raises(AttributeError):
            _where_to_add_values(four_by_four_raster, at, node_bc)


def test_where_to_add_values_link_bc(four_by_four_raster, at, link_bc):
    if at == "link":
        where = _where_to_add_values(four_by_four_raster, at, link_bc)
        assert len(where) == four_by_four_raster.number_of_links
    elif at == "node":
        with pytest.raises(ValueError):
            _where_to_add_values(four_by_four_raster, at, link_bc)
    else:
        with pytest.raises(AttributeError):
            _where_to_add_values(four_by_four_raster, at, link_bc)


def test_where_to_add_values_wrong_size(four_by_four_raster, at):
    where = np.full(four_by_four_raster.size(at) - 1, False)
    with pytest.raises(ValueError):
        _where_to_add_values(four_by_four_raster, at, where)


def test_where_to_add_values_with_bad_at(four_by_four_raster):
    with pytest.raises(GroupError):
        _where_to_add_values(four_by_four_raster, "not-a-place", None)


# test all grid elements


# where is size of elements
# where is list of types
# where is list but use patches (raises value error)


# x and y of multiple types of grid elements
# ask for a plane of something without X and y.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 21 16:52:10 2017

@author: margauxmouchene
"""

import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from landlab import RasterModelGrid
from landlab.components import (
    FlowAccumulator,
    FlowDirectorSteepest,
    TransportLengthHillslopeDiffuser,
)


def test_route_to_multiple_error_raised():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        TransportLengthHillslopeDiffuser(mg, erodibility=1.0, slope_crit=0.5)


def test_tl_hill_diff():
    """Test cases where S>Sc, S=Sc and S<Sc"""

    # Test cases where S>Sc, S=Sc and S<Sc
    # Set up a 3x16 grid with closed boundaries and initial elevations.
    mg = RasterModelGrid((3, 12))
    z = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            5.,
            1.9,
            1.9,
            1.9,
            1.9,
            1.3,
            1.3,
            1.3,
            1.3,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    mg.add_field("node", "topographic__elevation", z)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, True)

    # Parameter values for test
    k = 0.001
    Sc = 0.6

    # Instantiate flow director and tl hillslope diffuser
    fdir = FlowDirectorSteepest(mg)
    tl_diff = TransportLengthHillslopeDiffuser(mg, erodibility=k, slope_crit=Sc)

    # Run flow director
    fdir.run_one_step()

    # test slopes
    s_out = mg.at_node["topographic__steepest_slope"]
    s_test = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            3.1,
            0.,
            0.,
            0.,
            0.6,
            0.,
            0.,
            0.,
            0.3,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    assert_almost_equal(s_out, s_test, decimal=10)

    # Run tl hillslope diffusion component
    tl_diff.run_one_step(1.)

    # Test results
    # flux_out
    fo_test = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.025,
            0.,
            0.,
            0.,
            0.0006,
            0.,
            0.,
            0.,
            0.0003,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    fo_out = mg.at_node["sediment__flux_out"]
    assert_almost_equal(fo_out, fo_test, decimal=10)

    # updated elevation
    elev_test = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            4.975,
            1.9,
            1.9,
            1.9,
            1.8994,
            1.3,
            1.3,
            1.3,
            1.2997,
            1.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    elev_out = mg.at_node["topographic__elevation"]
    assert_almost_equal(elev_out, elev_test, decimal=10)

    # Run another time step because deposition and transfer were null
    # the first time
    fdir.run_one_step()
    tl_diff.run_one_step(1.)

    # Test results
    # flux_out
    fo_test = np.array(
        [
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            2.47500000e-02,
            0.00000000e+00,
            0.00000000e+00,
            6.00000000e-07,
            5.99400000e-04,
            0.00000000e+00,
            0.00000000e+00,
            3.00000000e-07,
            2.99700000e-04,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
            0.00000000e+00,
        ]
    )
    fo_out = mg.at_node["sediment__flux_out"]
    assert_almost_equal(fo_out, fo_test, decimal=10)

    # updated elevation
    elev_test = np.array(
        [
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            4.95025,
            1.925,
            1.9,
            1.8999994,
            1.8988006,
            1.3006,
            1.3,
            1.2999997,
            1.2994003,
            1.0003,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
            0.,
        ]
    )
    elev_out = mg.at_node["topographic__elevation"]
    assert_almost_equal(elev_out, elev_test, decimal=10)




import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid, VoronoiDelaunayGrid
from landlab.components import FlowAccumulator, FlowDirectorDINF
from landlab.components.flow_director import flow_direction_dinf


def test_not_implemented_voroni():
    x = [0, 0.1, 0.2, 0.3, 1, 1.1, 1.2, 1.3, 2, 2.1, 2.2, 2.3]
    y = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]
    vmg = VoronoiDelaunayGrid(x, y)
    with pytest.raises(NotImplementedError):
        flow_direction_dinf.flow_directions_dinf(vmg)


def test_D_infinity_low_closed_boundary_conditions():
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    z = np.array(
        [[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0], [0, 0, 0, 0]],
        dtype="float64",
    )
    mg.add_field("node", "topographic__elevation", z)
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    fd = FlowDirectorDINF(mg)
    fd.run_one_step()

    true_recievers = np.array(
        [
            [0, -1],
            [1, -1],
            [2, -1],
            [3, -1],
            [4, -1],
            [6, -1],
            [6, -1],
            [7, -1],
            [8, -1],
            [10, 6],
            [6, -1],
            [11, -1],
            [12, -1],
            [10, -1],
            [10, -1],
            [15, -1],
            [16, -1],
            [17, -1],
            [18, -1],
            [19, -1],
        ]
    )

    true_proportions = np.array(
        [
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [0.06058469, 0.93941531],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
        ]
    )
    assert_array_equal(fd.receivers, true_recievers)
    assert_array_equal(
        np.round(fd.proportions, decimals=6), np.round(true_proportions, decimals=6)
    )


def test_D_infinity_open_boundary_conditions():
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    z = mg.x_of_node + 2. * mg.y_of_node
    mg.add_field("node", "topographic__elevation", z)

    fd = FlowDirectorDINF(mg)
    fd.run_one_step()

    true_recievers = np.array(
        [
            [0, -1],
            [1, -1],
            [2, -1],
            [3, -1],
            [4, -1],
            [1, 0],
            [2, 1],
            [7, -1],
            [8, -1],
            [5, 4],
            [6, 5],
            [11, -1],
            [12, -1],
            [9, 8],
            [10, 9],
            [15, -1],
            [16, -1],
            [17, -1],
            [18, -1],
            [19, -1],
        ]
    )

    true_proportions = np.array(
        [
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [0.40966553, 0.59033447],
            [0.40966553, 0.59033447],
            [1., 0.],
            [1., 0.],
            [0.40966553, 0.59033447],
            [0.40966553, 0.59033447],
            [1., 0.],
            [1., 0.],
            [0.40966553, 0.59033447],
            [0.40966553, 0.59033447],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
            [1., 0.],
        ]
    )
    assert_array_equal(fd.receivers, true_recievers)
    assert_array_equal(
        np.round(fd.proportions, decimals=6), np.round(true_proportions, decimals=6)
    )


def test_D_infinity_flat():
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    mg.add_zeros("node", "topographic__elevation")

    fd = FlowDirectorDINF(mg)
    fd.run_one_step()

    node_ids = np.arange(mg.number_of_nodes)
    true_recievers = -1 * np.ones(fd.receivers.shape)
    true_recievers[:, 0] = node_ids

    true_proportions = np.zeros(fd.proportions.shape)
    true_proportions[:, 0] = 1

    assert_array_equal(fd.receivers, true_recievers)
    assert_array_equal(
        np.round(fd.proportions, decimals=6), np.round(true_proportions, decimals=6)
    )


def test_D_infinity_flat_closed_lower():
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    z = mg.add_zeros("node", "topographic__elevation")
    z[mg.core_nodes] += 1
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    fd = FlowDirectorDINF(mg)
    fd.run_one_step()

    node_ids = np.arange(mg.number_of_nodes)
    true_recievers = -1 * np.ones(fd.receivers.shape)
    true_recievers[:, 0] = node_ids

    true_proportions = np.zeros(fd.proportions.shape)
    true_proportions[:, 0] = 1

    assert_array_equal(fd.receivers, true_recievers)
    assert_array_equal(
        np.round(fd.proportions, decimals=6), np.round(true_proportions, decimals=6)
    )


def test_D_infinity_flat_closed_upper():
    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    z = mg.add_zeros("node", "topographic__elevation")
    z[mg.core_nodes] -= 1
    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    fd = FlowDirectorDINF(mg)
    fd.run_one_step()

    node_ids = np.arange(mg.number_of_nodes)
    true_recievers = -1 * np.ones(fd.receivers.shape)
    true_recievers[:, 0] = node_ids

    true_proportions = np.zeros(fd.proportions.shape)
    true_proportions[:, 0] = 1

    assert_array_equal(fd.receivers, true_recievers)
    assert_array_equal(
        np.round(fd.proportions, decimals=6), np.round(true_proportions, decimals=6)
    )


def test_D_infinity_SW_slope():
    mg = RasterModelGrid((10, 10))
    mg.add_field("topographic__elevation", mg.node_y + mg.node_x, at="node")
    fa = FlowAccumulator(mg, flow_director="FlowDirectorDINF")
    fa.run_one_step()

    # this one should all flow to the soutwest (third column of diagonal neighbors at node)
    node_ids = np.arange(mg.number_of_nodes)
    sw_diags = mg.diagonal_adjacent_nodes_at_node[:, 2]
    true_recievers = -1 * np.ones(fa.flow_director.receivers.shape)
    true_recievers[:, 0] = sw_diags
    true_recievers[mg.boundary_nodes, 0] = node_ids[mg.boundary_nodes]

    true_proportions = np.zeros(fa.flow_director.proportions.shape)
    true_proportions[:, 0] = 1

    assert_array_equal(true_recievers, fa.flow_director.receivers)
    assert_array_equal(true_proportions, fa.flow_director.proportions)


def test_D_infinity_WSW_slope():
    mg = RasterModelGrid((10, 10))
    mg.add_field(
        "topographic__elevation", mg.node_y * (2 ** 0.5 - 1.) + mg.node_x, at="node"
    )
    fa = FlowAccumulator(mg, flow_director="FlowDirectorDINF")
    fa.run_one_step()

    # this one should flow equally to west and southwest.
    node_ids = np.arange(mg.number_of_nodes)
    sw_diags = mg.diagonal_adjacent_nodes_at_node[:, 2]
    w_links = mg.adjacent_nodes_at_node[:, 2]
    true_recievers = -1 * np.ones(fa.flow_director.receivers.shape)
    true_recievers[mg.core_nodes, 0] = w_links[mg.core_nodes]
    true_recievers[mg.core_nodes, 1] = sw_diags[mg.core_nodes]
    true_recievers[mg.boundary_nodes, 0] = node_ids[mg.boundary_nodes]

    true_proportions = np.zeros(fa.flow_director.proportions.shape)
    true_proportions[mg.boundary_nodes, 0] = 1
    true_proportions[mg.core_nodes, 0] = 0.5
    true_proportions[mg.core_nodes, 1] = 0.5

    assert_array_equal(true_recievers, fa.flow_director.receivers)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_number_of_corners():
    """Test number of corners on a raster."""
    grid = RasterModelGrid((4, 5))
    assert grid.number_of_corners == 12


def test_add_ones_at_corners():
    """Test add a field to corners."""
    grid = RasterModelGrid((4, 5))
    grid.add_ones("z", at="corner")

    assert_array_equal(
        grid.at_corner["z"], [1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.]
    )


def test_add_field_at_corners():
    """Test add a field to corners."""
    grid = RasterModelGrid((4, 5))
    x = np.arange(grid.number_of_corners)
    grid.add_field("z", x, at="corner")

    assert_array_equal(grid.at_corner["z"], x)



"""Test Braun-Willett flow routing and fastscape stream power.

A simple driver implementing Braun-Willett flow routing and then the
fastscape stream power component.
"""
# DEJH, 09/15/14
import os

import numpy
from numpy.testing import assert_array_almost_equal

from landlab import ModelParameterDictionary, RasterModelGrid
from landlab.components import FlowAccumulator
from landlab.components.stream_power import FastscapeEroder as Fsc

_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_fastscape():
    input_str = os.path.join(_THIS_DIR, "drive_sp_params.txt")
    inputs = ModelParameterDictionary(input_str)
    nrows = inputs.read_int("nrows")
    ncols = inputs.read_int("ncols")
    dx = inputs.read_float("dx")
    dt = inputs.read_float("dt")
    time_to_run = inputs.read_float("run_time")
    uplift = inputs.read_float("uplift_rate")
    init_elev = inputs.read_float("init_elev")

    mg = RasterModelGrid(nrows, ncols, xy_spacing=dx)
    mg.set_closed_boundaries_at_grid_edges(False, False, True, True)

    mg.add_zeros("topographic__elevation", at="node")
    z = mg.zeros(at="node") + init_elev
    numpy.random.seed(0)
    mg["node"]["topographic__elevation"] = z + numpy.random.rand(len(z)) / 1000.

    fr = FlowAccumulator(mg, flow_director="D8")
    fsp = Fsc(mg, input_str, method="D8")
    elapsed_time = 0.
    while elapsed_time < time_to_run:
        if elapsed_time + dt > time_to_run:
            dt = time_to_run - elapsed_time
        mg = fr.run_one_step()
        mg = fsp.erode(mg, dt=dt)
        mg.at_node["topographic__elevation"][mg.core_nodes] += uplift * dt
        elapsed_time += dt

    z_trg = numpy.array(
        [
            5.48813504e-04,
            7.15189366e-04,
            6.02763376e-04,
            5.44883183e-04,
            4.23654799e-04,
            6.45894113e-04,
            1.01830760e-02,
            9.58036770e-03,
            6.55865452e-03,
            3.83441519e-04,
            7.91725038e-04,
            1.00142749e-02,
            8.80798884e-03,
            5.78387585e-03,
            7.10360582e-05,
            8.71292997e-05,
            9.81911417e-03,
            9.52243406e-03,
            7.55093226e-03,
            8.70012148e-04,
            9.78618342e-04,
            1.00629755e-02,
            8.49253798e-03,
            5.33216680e-03,
            1.18274426e-04,
            6.39921021e-04,
            9.88956320e-03,
            9.47119567e-03,
            6.43790696e-03,
            4.14661940e-04,
            2.64555612e-04,
            1.00450743e-02,
            8.37262908e-03,
            5.21540904e-03,
            1.87898004e-05,
            6.17635497e-04,
            9.21286940e-03,
            9.34022513e-03,
            7.51114450e-03,
            6.81820299e-04,
            3.59507901e-04,
            6.19166921e-03,
            7.10456176e-03,
            6.62585507e-03,
            6.66766715e-04,
            6.70637870e-04,
            2.10382561e-04,
            1.28926298e-04,
            3.15428351e-04,
            3.63710771e-04,
        ]
    )

    assert_array_almost_equal(mg.at_node["topographic__elevation"], z_trg)


def test_fastscape_new():
    input_str = os.path.join(_THIS_DIR, "drive_sp_params.txt")
    inputs = ModelParameterDictionary(input_str, auto_type=True)
    nrows = inputs.read_int("nrows")
    ncols = inputs.read_int("ncols")
    dx = inputs.read_float("dx")
    dt = inputs.read_float("dt")
    time_to_run = inputs.read_float("run_time")
    uplift = inputs.read_float("uplift_rate")
    init_elev = inputs.read_float("init_elev")

    mg = RasterModelGrid(nrows, ncols, xy_spacing=dx)
    mg.set_closed_boundaries_at_grid_edges(False, False, True, True)

    mg.add_zeros("topographic__elevation", at="node")
    z = mg.zeros(at="node") + init_elev
    numpy.random.seed(0)
    mg["node"]["topographic__elevation"] = z + numpy.random.rand(len(z)) / 1000.

    fr = FlowAccumulator(mg, flow_director="D8")
    fsp = Fsc(mg, **inputs)  # here's the diff from the above
    elapsed_time = 0.
    while elapsed_time < time_to_run:
        if elapsed_time + dt > time_to_run:
            dt = time_to_run - elapsed_time
        fr.run_one_step()
        fsp.run_one_step(dt)  # new style
        mg.at_node["topographic__elevation"][mg.core_nodes] += uplift * dt
        elapsed_time += dt

    z_trg = numpy.array(
        [
            5.48813504e-04,
            7.15189366e-04,
            6.02763376e-04,
            5.44883183e-04,
            4.23654799e-04,
            6.45894113e-04,
            1.01830760e-02,
            9.58036770e-03,
            6.55865452e-03,
            3.83441519e-04,
            7.91725038e-04,
            1.00142749e-02,
            8.80798884e-03,
            5.78387585e-03,
            7.10360582e-05,
            8.71292997e-05,
            9.81911417e-03,
            9.52243406e-03,
            7.55093226e-03,
            8.70012148e-04,
            9.78618342e-04,
            1.00629755e-02,
            8.49253798e-03,
            5.33216680e-03,
            1.18274426e-04,
            6.39921021e-04,
            9.88956320e-03,
            9.47119567e-03,
            6.43790696e-03,
            4.14661940e-04,
            2.64555612e-04,
            1.00450743e-02,
            8.37262908e-03,
            5.21540904e-03,
            1.87898004e-05,
            6.17635497e-04,
            9.21286940e-03,
            9.34022513e-03,
            7.51114450e-03,
            6.81820299e-04,
            3.59507901e-04,
            6.19166921e-03,
            7.10456176e-03,
            6.62585507e-03,
            6.66766715e-04,
            6.70637870e-04,
            2.10382561e-04,
            1.28926298e-04,
            3.15428351e-04,
            3.63710771e-04,
        ]
    )

    assert_array_almost_equal(mg.at_node["topographic__elevation"], z_trg)



"""General decorators for the landlab library.

General Landlab decorators
++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.utils.decorators.use_file_name_or_kwds
    ~landlab.utils.decorators.use_field_name_or_array
    ~landlab.utils.decorators.make_return_array_immutable
    ~landlab.utils.decorators.deprecated
"""

import inspect
import os
import textwrap
import warnings
from functools import wraps

import numpy as np
import six

from ..core.model_parameter_loader import load_params

try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec


class cache_result_in_object(object):
    def __init__(self, cache_as=None):
        self._attr = cache_as

    def __call__(self, func):
        name = self._attr or "_" + func.__name__

        @wraps(func)
        def _wrapped(obj):
            if not hasattr(obj, name):
                setattr(obj, name, func(obj))
            return getattr(obj, name)

        return _wrapped


class store_result_in_grid(object):
    def __init__(self, name=None):
        self._attr = name

    def __call__(self, func):
        @wraps(func)
        def _wrapped(grid):
            name = self._attr or "_" + func.__name__
            try:
                getattr(grid, name)
            except AttributeError:
                setattr(grid, name, func(grid))
            finally:
                return getattr(grid, name)

        return _wrapped


class store_result_in_dataset(object):
    def __init__(self, dataset=None, name=None):
        self._dataset = dataset
        self._attr = name

    def __call__(self, func):
        @wraps(func)
        def _wrapped(grid):
            name = self._attr or func.__name__
            if self._dataset:
                ds = getattr(grid, self._dataset)
            else:
                ds = grid

            if name not in ds:
                setattr(grid, self._dataset, ds.update(dict(name=func(grid))))

            return getattr(grid, self._dataset)[name].values

        return _wrapped


def read_only_array(func):
    """Decorate a function so that its return array is read-only.

    Parameters
    ----------
    func : function
        A function that returns a numpy array.

    Returns
    -------
    func
        A wrapped function that returns a read-only numpy array.
    """

    @wraps(func)
    def _wrapped(self, *args, **kwds):
        """Make the returned array read-only."""
        array = func(self, *args, **kwds)
        array.flags.writeable = False
        return array

    return _wrapped


def add_signature_to_doc(func):
    """Add a function signature to the first line of a docstring.

    Add the signature of a function to the first line of its docstring.
    This is useful for functions that are decorated in such a way
    that its signature changes.

    Parameters
    ----------
    func : function
        Function to add signature to.

    Returns
    -------
    str
        The new docstring with the function signature on the first line.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from landlab.utils.decorators import add_signature_to_doc

    >>> def foo(arg1, kwd=None):
    ...     '''Do something.'''
    ...     pass
    >>> print(add_signature_to_doc(foo))
    foo(arg1, kwd=None)
    <BLANKLINE>
    Do something.
    """
    argspec = getfullargspec(func)
    return """{name}{argspec}

{body}""".format(
        name=func.__name__,
        argspec=inspect.formatargspec(*argspec),
        body=inspect.getdoc(func),
    )


def use_file_name_or_kwds(func):

    """Decorate a method so that it takes a file name or keywords.

    Parameters
    ----------
    func : A function
        A method function that accepts a ModelGrid as a first argument.

    Returns
    -------
    function
        A function that takes an optional second argument, a file, from which
        to read keywords.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.utils.decorators import use_file_name_or_kwds

    >>> class MyClass(object):
    ...     @use_file_name_or_kwds
    ...     def __init__(self, grid, kw=0.):
    ...         self.kw = kw

    >>> grid = RasterModelGrid((4, 5))
    >>> foo = MyClass(grid)
    >>> foo.kw
    0.0

    >>> foo = MyClass(grid, "kw: 1945")
    >>> foo.kw
    1945
    >>> foo = MyClass(grid, "kw: 1945", kw=1973)
    >>> foo.kw
    1973

    >>> mpd = \"\"\"
    ... kw: kw value
    ... 1e6
    ... \"\"\"
    >>> foo = MyClass(grid, mpd)
    >>> foo.kw
    1000000.0
    """

    func.__doc__ = add_signature_to_doc(func)

    @wraps(func)
    def _wrapped(self, *args, **kwds):
        from ..grid import ModelGrid

        if not isinstance(args[0], ModelGrid):
            raise ValueError("first argument must be a ModelGrid")

        if len(args) == 2:
            warnings.warn(
                "Passing a file to a component's __init__ method is "
                "deprecated. Instead, pass parameters as keywords.",
                category=DeprecationWarning,
            )

            if os.path.isfile(args[1]):
                with open(args[1], "r") as fp:
                    params = load_params(fp)
            else:
                params = load_params(args[1])
        else:
            params = {}

        params.update(kwds)

        func(self, args[0], **params)

    return _wrapped


class use_field_name_or_array(object):

    """Decorate a function so that it accepts a field name or array.

    Parameters
    ----------
    func : function
        A function that accepts a grid and array as arguments.
    at_element : str
        The element type that the field is defined on ('node', 'cell',
        etc.)

    Returns
    -------
    func
        A wrapped function that accepts a grid and either a field name or
        a numpy array.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((4, 5), xy_spacing=(1, 2))

    >>> def my_func(grid, vals):
    ...     return grid.area_of_cell * vals
    >>> my_func(grid, np.arange(grid.number_of_cells))
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    Decorate the function so that the second argument can be array-like or
    the name of a field contained withing the grid. The decorator takes a
    single argument that is the name (as a `str`) of the grid element that
    the values are defined on ("node", "cell", etc.).

    >>> from landlab.utils.decorators import use_field_name_or_array
    >>> @use_field_name_or_array('cell')
    ... def my_func(grid, vals):
    ...     return grid.area_of_cell * vals

    The array of values now can be list or anything that can be converted to
    a numpy array.

    >>> my_func(grid, [0, 1, 2, 3, 4, 5])
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values doesn't have to be flat.

    >>> vals = np.array([[0, 1, 2], [3, 4, 5]])
    >>> my_func(grid, vals)
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values can be a field name.

    >>> _ = grid.add_field('cell', 'elevation', [0, 1, 2, 3, 4, 5])
    >>> my_func(grid, 'elevation')
    array([  0.,   2.,   4.,   6.,   8.,  10.])
    """

    def __init__(self, at_element):
        """Initialize the decorator.

        Parameters
        ----------
        at_element : str
            The element type that the field is defined on ('node', 'cell',
            etc.)
        """
        self._at = at_element

    def __call__(self, func):
        """Wrap the function."""

        func.__doc__ = add_signature_to_doc(func)

        @wraps(func)
        def _wrapped(grid, vals, *args, **kwds):
            """Convert the second argument to an array."""
            if isinstance(vals, six.string_types):
                vals = grid[self._at][vals]
            else:
                vals = np.asarray(vals).flatten()

            return func(grid, vals, *args, **kwds)

        return _wrapped


class use_field_name_array_or_value(object):

    """Decorate a function so that it accepts a field name, array, or value.

    Parameters
    ----------
    func : function
        A function that accepts a grid and array as arguments.
    at_element : str
        The element type that the field is defined on ('node', 'cell',
        etc.)

    Returns
    -------
    func
        A wrapped function that accepts a grid and either a field name,
        a numpy array, or a value as arguments.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((4, 5), xy_spacing=(1, 2))

    >>> def my_func(grid, vals):
    ...     return grid.area_of_cell * vals
    >>> my_func(grid, np.arange(grid.number_of_cells))
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    Decorate the function so that the second argument can be array-like,
    the name of a field contained withing the grid, or a value (float, int,
    etc.). The decorator takes a single argument that is the name (as a `str`)
    of the grid element that the values are defined on ("node", "cell", etc.).

    >>> from landlab.utils.decorators import use_field_name_array_or_value
    >>> @use_field_name_array_or_value('cell')
    ... def my_func(grid, vals):
    ...     return grid.area_of_cell * vals

    The array of values now can be list or anything that can be converted to
    a numpy array.

    >>> my_func(grid, [0, 1, 2, 3, 4, 5])
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values doesn't have to be flat.

    >>> vals = np.array([[0, 1, 2], [3, 4, 5]])
    >>> my_func(grid, vals)
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values can be a field name.

    >>> _ = grid.add_field('cell', 'elevation', [0, 1, 2, 3, 4, 5])
    >>> my_func(grid, 'elevation')
    array([  0.,   2.,   4.,   6.,   8.,  10.])

    The array of values can be a value (float, int, etc.).

    >>> my_func(grid, 4.0)
    array([ 8.,  8.,  8.,  8.,  8.,  8.])
    """

    def __init__(self, at_element):
        """Initialize the decorator.

        Parameters
        ----------
        at_element : str
            The element type that the field is defined on ('node', 'cell',
            etc.)
        """
        self._at = at_element

    def __call__(self, func):
        """Wrap the function."""

        func.__doc__ = add_signature_to_doc(func)

        @wraps(func)
        def _wrapped(grid, vals, *args, **kwds):
            """Convert the second argument to an array."""
            if isinstance(vals, six.string_types):
                vals = grid[self._at][vals]
            else:
                expected_size = grid.size(self._at)
                vals = np.asarray(vals).flatten()
                if vals.size == 1:
                    vals = np.broadcast_to(vals, (expected_size,))

                if vals.size != expected_size:
                    raise ValueError(
                        (
                            "Array passed to function decorated with "
                            "use_field_name_array_or_value is not "
                            "the size of fields at " + self._at
                        )
                    )
            return func(grid, vals, *args, **kwds)

        return _wrapped


def make_return_array_immutable(func):
    """Decorate a function so that its return array is read-only.

    Parameters
    ----------
    func : function
        A function that returns a numpy array.

    Returns
    -------
    func
        A wrapped function that returns a read-only view of an array.
    """

    @wraps(func)
    def _wrapped(self, *args, **kwds):
        """Make the returned array read-only."""
        array = func(self, *args, **kwds)
        immutable_array = array.view()
        immutable_array.flags.writeable = False
        return immutable_array

    return _wrapped


# def deprecated(use, version):
#     """Mark a function as deprecated.
#
#     Parameters
#     ----------
#     use : str
#         Name of replacement function to use.
#     version : str
#         Version number when function was marked as deprecated.
#
#     Returns
#     -------
#     func
#         A wrapped function that issues a deprecation warning.
#     """
#     def real_decorator(func):
#
#         def _wrapped(*args, **kwargs):
#             """Warn that the function is deprecated before calling it."""
#             warnings.warn(
#                 "Call to deprecated function {name}.".format(
#                     name=func.__name__), category=DeprecationWarning)
#             return func(*args, **kwargs)
#         _wrapped.__dict__.update(func.__dict__)
#
#         return _wrapped
#
#     return real_decorator


def deprecated(use, version):
    """Mark a function as deprecated.

    Parameters
    ----------
    use : str
        Name of replacement function to use.
    version : str
        Version number when function was marked as deprecated.

    Returns
    -------
    func
        A wrapped function that issues a deprecation warning.
    """

    def real_decorator(func):
        warning_str = """
.. note:: This method is deprecated as of Landlab version {ver}.

    Use :func:`{use}` instead.

""".format(
            ver=version, use=use
        )

        doc_lines = (func.__doc__ or "").split(os.linesep)

        for lineno, line in enumerate(doc_lines):
            if len(line.rstrip()) == 0:
                break

        head = doc_lines[:lineno]
        body = doc_lines[lineno:]

        head = textwrap.dedent(os.linesep.join(head))
        body = textwrap.dedent(os.linesep.join(body))

        func.__doc__ = os.linesep.join([head, warning_str, body])

        @wraps(func)
        def _wrapped(*args, **kwargs):
            if func.__name__.startswith("_"):
                pass
            else:
                warnings.warn(
                    message="Call to deprecated function {name}.".format(
                        name=func.__name__
                    ),
                    category=DeprecationWarning,
                )
            return func(*args, **kwargs)

        _wrapped.__dict__.update(func.__dict__)

        return _wrapped

    return real_decorator



#! /usr/bin/env python
"""Read data from a NetCDF file into a RasterModelGrid.

Read netcdf
+++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.io.netcdf.read.read_netcdf
"""

try:
    import netCDF4 as nc4
except ImportError:
    import warnings

    warnings.warn("Unable to import netCDF4.", ImportWarning)

import numpy as np
from scipy.io import netcdf as nc

from landlab.io import (
    MismatchGridDataSizeError,
    MismatchGridXYLowerLeft,
    MismatchGridXYSpacing,
)
from landlab.io.netcdf._constants import (
    _AXIS_COORDINATE_NAMES,
    _AXIS_DIMENSION_NAMES,
    _COORDINATE_NAMES,
)
from landlab.io.netcdf.errors import NotRasterGridError
from landlab.utils import add_halo


def _length_of_axis_dimension(root, axis_name):
    """Get the size of an axis by axis name.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF object.
    axis_name : str
        Name of the axis in the NetCDF file.

    Returns
    -------
    int
        Size of the dimension.
    """
    try:
        return len(root.dimensions[axis_name])
    except TypeError:
        return root.dimensions[axis_name]


def _read_netcdf_grid_shape(root):
    """Get the shape of a raster grid.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF object.

    Returns
    -------
    (rows, columns)
        Shape of the grid.
    """
    shape = []
    for axis_name in _AXIS_DIMENSION_NAMES:
        try:
            shape.append(_length_of_axis_dimension(root, axis_name))
        except KeyError:
            pass
    return tuple(shape)


def _read_netcdf_coordinate_values(root):
    """Get arrays of coordinates for grid points.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of ndarray
        Node coordinates for each dimension.
    """
    values = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            values.append(root.variables[coordinate_name][:].copy())
        except KeyError:
            pass
    return tuple(values)


def _read_netcdf_coordinate_units(root):
    """Get units for coodinate values.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of str
        Units for each coordinate.
    """
    units = []
    for coordinate_name in _AXIS_COORDINATE_NAMES:
        try:
            units.append(root.variables[coordinate_name].units)
        except KeyError:
            pass
    return tuple(units)


def _read_netcdf_structured_grid(root):
    """Get node coordinates for a structured grid.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of ndarray
        Node coordinates for each dimension reshaped to match the shape
        of the grid.
    """
    shape = _read_netcdf_grid_shape(root)
    coordinates = _read_netcdf_coordinate_values(root)

    for coordinate in coordinates:
        coordinate.shape = shape

    return coordinates


def _read_netcdf_raster_structured_grid(root):
    """Get node coordinates for a structured grid written as a raster.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    tuple of ndarray
        Node coordinates for each dimension reshaped to match the shape
        of the grid.
    """
    shape = _read_netcdf_grid_shape(root)
    coordinates = _read_netcdf_coordinate_values(root)

    if len(coordinates) != 2:
        assert ValueError("Rasters must have only two spatial coordinate dimensions")
    else:
        coordinates = np.meshgrid(coordinates[0], coordinates[1], indexing="ij")

    for coordinate in coordinates:
        coordinate.shape = shape

    return coordinates


def _read_netcdf_structured_data(root):
    """Get data values for a structured grid.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF file.

    Returns
    -------
    dict
        Data values, reshaped to match that of the grid. Keys are the
        variable names as read from the NetCDF file.
    """
    fields = dict()
    grid_mapping_exists = False
    grid_mapping_dict = None
    for (name, var) in root.variables.items():
        # identify if a grid mapping variable exist and do not pass it as a field
        if name not in _COORDINATE_NAMES:
            if hasattr(var, "grid_mapping"):
                grid_mapping = getattr(var, "grid_mapping")
                if type(grid_mapping) is bytes:
                    grid_mapping = grid_mapping.decode("utf-8")
                grid_mapping_exists = True

    dont_use = list(_COORDINATE_NAMES)
    if grid_mapping_exists:
        dont_use.append(grid_mapping)
    for (name, var) in root.variables.items():
        if name not in dont_use:
            fields[name] = var[:].copy()
            fields[name].shape = (fields[name].size,)

    if grid_mapping_exists:
        grid_mapping_variable = root.variables[grid_mapping]
        grid_mapping_dict = {"name": grid_mapping}
        try:
            for att in grid_mapping_variable.ncattrs():
                grid_mapping_dict[att] = getattr(grid_mapping_variable, att)
        except AttributeError:  # if scipy is doing the reading
            for att in var._attributes:
                grid_mapping_dict[att] = getattr(grid_mapping_variable, att)
    return fields, grid_mapping_dict


def _get_raster_spacing(coords):
    """Get the row and column spacing of a raster grid.

    Parameters
    ----------
    root : netcdf_file
        A NetCDF object.

    Returns
    -------
    (dy, dx)
        Spacing of grid rows and columns.
    """
    spacing = np.empty(len(coords), dtype=np.float64)

    for (axis, coord) in enumerate(coords):
        coord_spacing = np.diff(coord, axis=axis)
        if not np.all(coord_spacing == coord_spacing.flat[0]):
            raise NotRasterGridError()
        spacing[axis] = coord_spacing.flat[0]

    if not np.all(spacing == spacing[0]):
        raise NotRasterGridError()

    return spacing[0]


def read_netcdf(
    nc_file, grid=None, name=None, just_grid=False, halo=0, nodata_value=-9999.0
):
    """Create a :class:`~.RasterModelGrid` from a netcdf file.

    Create a new :class:`~.RasterModelGrid` from the netcdf file, *nc_file*.
    If the netcdf file also contains data, add that data to the grid's fields.
    To create a new grid without any associated data from the netcdf file,
    set the *just_grid* keyword to ``True``.

    A halo can be added with the keyword *halo*.

    If you want the fields to be added to an existing grid, it can be passed
    to the keyword argument *grid*.

    Parameters
    ----------
    nc_file : str
        Name of a netcdf file.
    grid : *grid* , optional
        Adds data to an existing *grid* instead of creating a new one.
    name : str, optional
        Add only fields with NetCDF variable name to the grid. Default is to
        add all NetCDF varibles to the grid.
    just_grid : boolean, optional
        Create a new grid but don't add value data.
    halo : integer, optional
        Adds outer border of depth halo to the *grid*.
    nodata_value : float, optional
        Value that indicates an invalid value. Default is -9999.

    Returns
    -------
    :class:`~.RasterModelGrid`
        A newly-created :class:`~.RasterModelGrid`.

    Examples
    --------
    Import :func:`read_netcdf` and the path to an example netcdf file included
    with landlab.

    >>> from landlab.io.netcdf import read_netcdf
    >>> from landlab.io.netcdf import NETCDF4_EXAMPLE_FILE

    Create a new grid from the netcdf file. The example grid is a uniform
    rectilinear grid with 4 rows and 3 columns of nodes with unit spacing.
    The data file also contains data defined at the nodes for the grid for
    a variable called, *surface__elevation*.

    >>> grid = read_netcdf(NETCDF4_EXAMPLE_FILE)
    >>> grid.shape == (4, 3)
    True
    >>> grid.dy, grid.dx
    (1.0, 1.0)
    >>> list(grid.at_node.keys())
    ['surface__elevation']
    >>> grid.at_node['surface__elevation']
    array([  0.,   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.,
            11.])

    :func:`read_netcdf` will try to determine the format of the netcdf file.
    For example, the same call will also work for *netcdf3*-formatted files.

    >>> from landlab.io.netcdf import NETCDF3_64BIT_EXAMPLE_FILE
    >>> grid = read_netcdf(NETCDF3_64BIT_EXAMPLE_FILE)
    >>> grid.shape == (4, 3)
    True
    >>> grid.dy, grid.dx
    (1.0, 1.0)

    A more complicated example might add data with a halo to an existing grid.
    Note that the lower left corner must be specified correctly for the data
    and the grid to align correctly.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((6, 5), xy_of_lower_left=(-1., -1.))
    >>> grid = read_netcdf(
    ...     NETCDF4_EXAMPLE_FILE,
    ...     grid=grid,
    ...     halo=1,
    ...     nodata_value=-1)
    >>> grid.at_node['surface__elevation'].reshape(grid.shape)
    array([[ -1.,  -1.,  -1.,  -1.,  -1.],
           [ -1.,   0.,   1.,   2.,  -1.],
           [ -1.,   3.,   4.,   5.,  -1.],
           [ -1.,   6.,   7.,   8.,  -1.],
           [ -1.,   9.,  10.,  11.,  -1.],
           [ -1.,  -1.,  -1.,  -1.,  -1.]])
    """
    from landlab import RasterModelGrid

    try:
        root = nc.netcdf_file(nc_file, "r", version=2)
    except TypeError:
        root = nc4.Dataset(nc_file, "r", format="NETCDF4")

    try:
        node_coords = _read_netcdf_structured_grid(root)
    except ValueError:
        if (len(root.variables["x"].dimensions) == 1) and (
            len(root.variables["y"].dimensions) == 1
        ):

            node_coords = _read_netcdf_raster_structured_grid(root)
        else:
            assert ValueError(
                "x and y dimensions must both either be 2D "
                "(nj, ni) or 1D (ni,) and (nj)."
            )

    assert len(node_coords) == 2

    dx = _get_raster_spacing(node_coords)
    xy_spacing = (dx, dx)
    shape = node_coords[0].shape
    xy_of_lower_left = (
        node_coords[0].min() - halo * dx,
        node_coords[1].min() - halo * dx,
    )

    if grid is not None:
        if (grid.number_of_node_rows != shape[0] + 2 * halo) or (
            grid.number_of_node_columns != shape[1] + 2 * halo
        ):
            raise MismatchGridDataSizeError(
                shape[0] + 2 * halo * shape[1] + 2 * halo,
                grid.number_of_node_rows * grid.number_of_node_columns,
            )
        if (grid.dx, grid.dy) != xy_spacing:
            raise MismatchGridXYSpacing((grid.dx, grid.dy), xy_spacing)

        if grid.xy_of_lower_left != xy_of_lower_left:
            raise MismatchGridXYLowerLeft(grid.xy_of_lower_left, xy_of_lower_left)

    if grid is None:
        grid = RasterModelGrid(
            shape, xy_spacing=xy_spacing, xy_of_lower_left=xy_of_lower_left
        )

    if not just_grid:
        fields, grid_mapping_dict = _read_netcdf_structured_data(root)
        for (field_name, values) in fields.items():

            # add halo if necessary
            if halo > 0:
                values = add_halo(
                    values.reshape(shape), halo=halo, halo_value=nodata_value
                ).reshape((-1,))

            # add only the requested fields.
            if (name is None) or (field_name == name):
                add_field = True
            else:
                add_field = False

            if add_field:
                grid.add_field(field_name, values, at="node", noclobber=False)

        if (name is not None) and (name not in grid.at_node):
            raise ValueError(
                "Specified field {name} was not in provided NetCDF.".format(name=name)
            )

    # save grid mapping
    if grid_mapping_dict is not None:
        grid.grid_mapping = grid_mapping_dict

    root.close()

    return grid



import numpy as np
from numpy.testing import assert_array_equal

import landlab.utils.structured_grid as sgrid


def test_default():
    neighbors = sgrid.neighbor_node_array((2, 3))

    X = sgrid.BAD_INDEX_VALUE
    assert_array_equal(
        neighbors,
        np.array(
            [
                [1, 3, X, X],
                [2, 4, 0, X],
                [X, 5, 1, X],
                [4, X, X, 0],
                [5, X, 3, 1],
                [X, X, 4, 2],
            ]
        ).T,
    )

    assert neighbors.flags["C_CONTIGUOUS"]
    assert neighbors.base is None


def test_set_out_of_bounds():
    neighbors = sgrid.neighbor_node_array((2, 3), inactive=-1)
    assert_array_equal(
        neighbors,
        np.array(
            [
                [1, 3, -1, -1],
                [2, 4, 0, -1],
                [-1, 5, 1, -1],
                [4, -1, -1, 0],
                [5, -1, 3, 1],
                [-1, -1, 4, 2],
            ]
        ).T,
    )


def test_boundary_node_mask_no_actives():
    neighbors = sgrid.neighbor_node_array(
        (2, 3), inactive=-2, closed_boundary_nodes=np.arange(6)
    )
    assert_array_equal(neighbors, -2 * np.ones((6, 4)).T)


def test_boundary_node_mask_3x3():
    neighbors = sgrid.neighbor_node_array(
        (3, 3), inactive=-2, closed_boundary_nodes=[0, 1, 2, 3, 5, 6, 7, 8]
    )
    assert_array_equal(neighbors, -2 * np.ones((9, 4)).T)


def test_boundary_node_mask():
    X = -1
    neighbors = sgrid.neighbor_node_array(
        (3, 3), inactive=X, closed_boundary_nodes=[0, 2, 3, 5, 6, 7, 8]
    )
    expected = np.array(
        [
            [X, X, X, X],
            [X, 4, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, 1],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors, expected.T)


def test_open_boundary():
    X = -1
    neighbors = sgrid.neighbor_node_array(
        (3, 3),
        closed_boundary_nodes=[3, 5, 6, 7, 8],
        open_boundary_nodes=[0, 1, 2],
        inactive=X,
    )
    expected = np.array(
        [
            [X, X, X, X],
            [X, 4, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, 1],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors.T, expected)


def test_open_boundary_4x5():
    X = -1
    neighbors = sgrid.neighbor_node_array(
        (4, 5),
        closed_boundary_nodes=[5, 10, 15, 16, 17, 18, 19],
        open_boundary_nodes=[0, 1, 2, 3, 4, 9, 14],
        inactive=X,
    )
    expected = np.array(
        [
            [X, X, X, X],
            [X, 6, X, X],
            [X, 7, X, X],
            [X, 8, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [7, 11, X, 1],
            [8, 12, 6, 2],
            [9, 13, 7, 3],
            [X, X, 8, X],
            [X, X, X, X],
            [12, X, X, 6],
            [13, X, 11, 7],
            [14, X, 12, 8],
            [X, X, 13, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors.T, expected)


def test_open_boundary_on_perimeter():
    X = -1

    neighbors = sgrid.neighbor_node_array(
        (5, 4), open_boundary_nodes=sgrid.boundary_nodes((5, 4)), inactive=X
    )

    expected = np.array(
        [
            [X, X, X, X],
            [X, 5, X, X],
            [X, 6, X, X],
            [X, X, X, X],
            [5, X, X, X],
            [6, 9, 4, 1],
            [7, 10, 5, 2],
            [X, X, 6, X],
            [9, X, X, X],
            [10, 13, 8, 5],
            [11, 14, 9, 6],
            [X, X, 10, X],
            [13, X, X, X],
            [14, 17, 12, 9],
            [15, 18, 13, 10],
            [X, X, 14, X],
            [X, X, X, X],
            [X, X, X, 13],
            [X, X, X, 14],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors, expected.T)


def test_closed_boundary_on_perimeter():
    X = -1

    neighbors = sgrid.neighbor_node_array(
        (5, 4), closed_boundary_nodes=sgrid.boundary_nodes((5, 4)), inactive=X
    )

    expected = np.array(
        [
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [6, 9, X, X],
            [X, 10, 5, X],
            [X, X, X, X],
            [X, X, X, X],
            [10, 13, X, 5],
            [X, 14, 9, 6],
            [X, X, X, X],
            [X, X, X, X],
            [14, X, X, 9],
            [X, X, 13, 10],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors, expected.T)


def test_closed_boundaries_in_middle():
    X = -1

    neighbors = sgrid.neighbor_node_array(
        (5, 4), closed_boundary_nodes=sgrid.boundary_nodes((5, 4)), inactive=X
    )

    expected = np.array(
        [
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [6, 9, X, X],
            [X, 10, 5, X],
            [X, X, X, X],
            [X, X, X, X],
            [10, 13, X, 5],
            [X, 14, 9, 6],
            [X, X, X, X],
            [X, X, X, X],
            [14, X, X, 9],
            [X, X, 13, 10],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
            [X, X, X, X],
        ]
    )
    assert_array_equal(neighbors, expected.T)



from .chi_index import ChiFinder
from .diffusion import LinearDiffuser
from .fire_generator import FireGenerator
from .detachment_ltd_erosion import DetachmentLtdErosion
from .detachment_ltd_erosion import DepthSlopeProductErosion
from .flexure import Flexure
from .flow_routing import FlowRouter, DepressionFinderAndRouter
from .nonlinear_diffusion import PerronNLDiffuse
from .flow_director import FlowDirectorD8
from .flow_director import FlowDirectorSteepest
from .flow_director import FlowDirectorMFD
from .flow_director import FlowDirectorDINF
from .flow_accum import FlowAccumulator
from .flow_accum import LossyFlowAccumulator
from .overland_flow import OverlandFlowBates, OverlandFlow
from .overland_flow import KinwaveImplicitOverlandFlow
from .overland_flow import KinwaveOverlandFlowModel
from .potentiality_flowrouting import PotentialityFlowRouter
from .pet import PotentialEvapotranspiration
from .radiation import Radiation
from .soil_moisture import SoilMoisture
from .vegetation_dynamics import Vegetation
from .lake_fill import LakeMapperBarnes
from .sink_fill import SinkFiller, SinkFillerBarnes
from .steepness_index import SteepnessFinder
from .stream_power import (
    StreamPowerEroder,
    FastscapeEroder,
    StreamPowerSmoothThresholdEroder,
    SedDepEroder,
)
from .uniform_precip import PrecipitationDistribution
from .spatial_precip import SpatialPrecipitationDistribution
from .soil_moisture import SoilInfiltrationGreenAmpt
from .plant_competition_ca import VegCA
from .gflex import gFlex
from .drainage_density import DrainageDensity
from .weathering import ExponentialWeatherer
from .depth_dependent_diffusion import DepthDependentDiffuser
from .taylor_nonlinear_hillslope_flux import TaylorNonLinearDiffuser
from .depth_dependent_taylor_soil_creep import DepthDependentTaylorDiffuser
from .erosion_deposition import ErosionDeposition
from .space import Space
from .landslides import LandslideProbability
from .transport_length_diffusion import TransportLengthHillslopeDiffuser
from .normal_fault import NormalFault
from .lithology import Lithology, LithoLayers

COMPONENTS = [
    ChiFinder,
    LinearDiffuser,
    Flexure,
    FlowRouter,
    DepressionFinderAndRouter,
    PerronNLDiffuse,
    OverlandFlowBates,
    OverlandFlow,
    KinwaveImplicitOverlandFlow,
    KinwaveOverlandFlowModel,
    PotentialEvapotranspiration,
    PotentialityFlowRouter,
    Radiation,
    SinkFiller,
    SinkFillerBarnes,
    LakeMapperBarnes,
    StreamPowerEroder,
    StreamPowerSmoothThresholdEroder,
    FastscapeEroder,
    SedDepEroder,
    PrecipitationDistribution,
    SpatialPrecipitationDistribution,
    SteepnessFinder,
    DetachmentLtdErosion,
    gFlex,
    SoilInfiltrationGreenAmpt,
    FireGenerator,
    SoilMoisture,
    Vegetation,
    VegCA,
    DrainageDensity,
    ExponentialWeatherer,
    DepthDependentDiffuser,
    TaylorNonLinearDiffuser,
    DepthSlopeProductErosion,
    FlowDirectorD8,
    FlowDirectorSteepest,
    FlowDirectorMFD,
    FlowDirectorDINF,
    FlowAccumulator,
    LossyFlowAccumulator,
    Space,
    ErosionDeposition,
    LandslideProbability,
    DepthDependentTaylorDiffuser,
    NormalFault,
    Lithology,
    LithoLayers,
    TransportLengthHillslopeDiffuser,
]

__all__ = [cls.__name__ for cls in COMPONENTS]



#! /usr/bin/env

# import landlab.utils.count_repeats
# from landlab.utils.count_repeats import count_repeats
from .count_repeats import count_repeated_values
from .source_tracking_algorithm import (
    track_source,
    convert_arc_flow_directions_to_landlab_node_ids,
    find_unique_upstream_hsd_ids_and_fractions,
)
from .watershed import (
    get_watershed_mask,
    get_watershed_masks_with_area_threshold,
    get_watershed_nodes,
    get_watershed_outlet,
    get_watershed_masks,
)
from .stable_priority_queue import StablePriorityQueue

from .add_halo import add_halo

__all__ = [
    "add_halo",
    "count_repeated_values",
    "track_source",
    "convert_arc_flow_directions_to_landlab_node_ids",
    "find_unique_upstream_hsd_ids_and_fractions",
    "get_watershed_mask",
    "get_watershed_masks_with_area_threshold",
    "get_watershed_nodes",
    "get_watershed_outlet",
    "get_watershed_masks",
    "StablePriorityQueue",
]



import pytest

from landlab import RasterModelGrid
from landlab.components.soil_moisture import SoilInfiltrationGreenAmpt
from landlab.components.soil_moisture.soil_moisture_dynamics import SoilMoisture


@pytest.fixture
def sm():
    grid = RasterModelGrid((20, 20), xy_spacing=10e0)
    grid.add_zeros("vegetation__plant_functional_type", at="cell", dtype=int)
    return SoilMoisture(grid)


@pytest.fixture
def si():
    grid = RasterModelGrid((10, 10), xy_spacing=25)
    grid.add_ones("soil_water_infiltration__depth", at="node", dtype=float)
    grid.add_ones("surface_water__depth", at="node")
    hydraulic_conductivity = 2.5 * (10 ** -5)
    grid.at_node["surface_water__depth"] *= 0.5
    grid.at_node["soil_water_infiltration__depth"] *= 10 ** -5
    return SoilInfiltrationGreenAmpt(
        grid,
        hydraulic_conductivity=hydraulic_conductivity,
        soil_bulk_density=1700.,
        rock_density=2650.,
        initial_soil_moisture_content=0.2,
        soil_type="silt loam",
        volume_fraction_coarse_fragments=0.6,
        coarse_sed_flag=False,
        surface_water_minimum_depth=1.e-7,
        soil_pore_size_distribution_index=None,
        soil_bubbling_pressure=None,
        wetting_front_capillary_pressure_head=None,
    )



"""
Unit tests for \
    landlab.components.vegetation_dynamics.vegetation_dynamics
"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(veg):
    assert veg.name == "Vegetation"


def test_input_var_names(veg):
    assert sorted(veg.input_var_names) == [
        "surface__evapotranspiration",
        "surface__potential_evapotranspiration_30day_mean",
        "surface__potential_evapotranspiration_rate",
        "vegetation__plant_functional_type",
        "vegetation__water_stress",
    ]


def test_output_var_names(veg):
    assert sorted(veg.output_var_names) == [
        "vegetation__cover_fraction",
        "vegetation__dead_biomass",
        "vegetation__dead_leaf_area_index",
        "vegetation__live_biomass",
        "vegetation__live_leaf_area_index",
    ]


def test_var_units(veg):
    assert set(veg.input_var_names) | set(veg.output_var_names) == set(
        dict(veg.units).keys()
    )

    assert veg.var_units("vegetation__live_leaf_area_index") == "None"
    assert veg.var_units("vegetation__dead_leaf_area_index") == "None"
    assert veg.var_units("vegetation__cover_fraction") == "None"
    assert veg.var_units("surface__evapotranspiration") == "mm"
    assert veg.var_units("surface__potential_evapotranspiration_rate") == "mm"
    assert veg.var_units("surface__potential_evapotranspiration_30day_mean") == "mm"
    assert veg.var_units("vegetation__water_stress") == "None"
    assert veg.var_units("vegetation__live_biomass") == "g m^-2 d^-1"
    assert veg.var_units("vegetation__dead_biomass") == "g m^-2 d^-1"
    assert veg.var_units("vegetation__plant_functional_type") == "None"


def test_grid_shape(veg):
    assert veg.grid.number_of_node_rows == _SHAPE[0]
    assert veg.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(veg):
    assert veg.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(veg):
    assert veg.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(veg):
    for name in veg.grid["node"]:
        field = veg.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            veg.grid.number_of_node_rows * veg.grid.number_of_node_columns,
        )

    for name in veg.grid["cell"]:
        field = veg.grid["cell"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            veg.grid.number_of_cell_rows * veg.grid.number_of_cell_columns,
        )

    with pytest.raises(KeyError):
        veg.grid["not_a_var_name"]


def test_field_initialized_to_zero(veg):
    for name in veg.grid["node"]:
        field = veg.grid["node"][name]
        assert_array_almost_equal(field, np.zeros(veg.grid.number_of_nodes))
    for name in veg.grid["cell"]:
        field = veg.grid["cell"][name]
        assert_array_almost_equal(field, np.zeros(veg.grid.number_of_cells))
